"""
Test suite for multi_index.py
"""
import numpy as np
import pytest
from conftest import (
    LpDegree,
    PolyDegree,
    SpatialDimension,
    assert_call,
    assert_multi_index_equal,
)
from numpy.testing import assert_, assert_equal, assert_raises

from minterpy import MultiIndexSet
from minterpy.core.utils import (
    find_match_between,
    get_exponent_matrix,
    _get_poly_degree,
    is_lexicographically_complete,
)

# test initialization
def test_init_from_exponents(SpatialDimension, PolyDegree, LpDegree):
    exponents = get_exponent_matrix(SpatialDimension, PolyDegree, LpDegree)
    assert_call(MultiIndexSet, exponents)
    assert_call(MultiIndexSet, exponents, lp_degree=LpDegree)


def test_init_fail_from_exponents():
    exponents = get_exponent_matrix(2, 1, 1)
    exponents[0] = 1
    assert_raises(ValueError, MultiIndexSet, exponents)


def test_init_from_degree(SpatialDimension, PolyDegree, LpDegree):
    assert_call(MultiIndexSet.from_degree, SpatialDimension, PolyDegree)
    assert_call(
        MultiIndexSet.from_degree, SpatialDimension, PolyDegree, lp_degree=LpDegree
    )
    multi_index = MultiIndexSet.from_degree(SpatialDimension, PolyDegree, LpDegree)

    exponents = get_exponent_matrix(SpatialDimension, PolyDegree, LpDegree)
    groundtruth = MultiIndexSet(exponents, lp_degree=LpDegree)
    assert_multi_index_equal(multi_index, groundtruth)


def test_init_fail_from_degree():
    assert_raises(TypeError, MultiIndexSet.from_degree, 1.0, 1)
    assert_raises(TypeError, MultiIndexSet.from_degree, 1, 1.0)

# Test methods

def test_add_exponents(SpatialDimension, PolyDegree, LpDegree):
    """Test the add_exponents method of a MultiIndex instance.
    
    Notes
    -----
    - This is related to the fix for Issue #75.
    """
    
    # Create 2 exponents, one twice the polynomial degree of the other
    exponents_1 = get_exponent_matrix(SpatialDimension, PolyDegree, LpDegree)
    exponents_2 = get_exponent_matrix(SpatialDimension, 2*PolyDegree, LpDegree)

    # Compute the set difference between the larger set and the smaller set
    exponents_diff = np.array(
        list(set(map(tuple, exponents_2)) - set(map(tuple, exponents_1)))
    )

    # Create the multi-index sets
    mi_1 = MultiIndexSet(exponents_1, lp_degree=LpDegree)
    mi_2 = MultiIndexSet(exponents_2, lp_degree=LpDegree)

    # Act: Add the exponents difference to the smaller multi-index set
    mi_added = mi_1.add_exponents(exponents_diff)

    # Assert: The added multi-index must be the same as the big one
    assert_multi_index_equal(mi_added, mi_2)

# test attributes


def test_attributes(SpatialDimension, PolyDegree, LpDegree):
    exponents = get_exponent_matrix(SpatialDimension, PolyDegree, LpDegree)
    multi_index = MultiIndexSet(exponents, lp_degree=LpDegree)
    assert_(isinstance(multi_index, MultiIndexSet))
    assert_equal(exponents, multi_index.exponents)
    assert_(multi_index.lp_degree == LpDegree)
    assert_(multi_index.poly_degree == PolyDegree)

    number_of_monomials, dim = exponents.shape
    assert_(len(multi_index) == number_of_monomials)
    assert_(multi_index.spatial_dimension == dim)


def test_attributes_incomplete_exponents(SpatialDimension, PolyDegree, LpDegree):
    """Test the attributes with an incomplete exponents for MultiIndexSet.
    
    Notes
    -----
    - This is related to the fix for Issue #66.
    """
    exponents = get_exponent_matrix(SpatialDimension, PolyDegree, LpDegree)

    # Create an incomplete multi-index set
    # by adding a new exponent of higher degree
    exponents_incomplete = np.insert(
        exponents, len(exponents), exponents[-1]+2, axis=0)
    multi_index_incomplete = MultiIndexSet(
        exponents_incomplete, lp_degree=LpDegree)

    # Make sure the incomplete exponents are indeed incomplete
    assert_(not is_lexicographically_complete(exponents_incomplete))

    # Compute the reference polynomial degree given the multi-index set    
    poly_degree = _get_poly_degree(exponents_incomplete, LpDegree)

    # Assertion of attributes
    assert_equal(exponents_incomplete, multi_index_incomplete.exponents)
    assert_(multi_index_incomplete.lp_degree == LpDegree)
    assert_(multi_index_incomplete.poly_degree == poly_degree)

    num_of_monomials_incomplete, dim_incomplete = exponents_incomplete.shape
    assert_(len(multi_index_incomplete) == num_of_monomials_incomplete)
    assert_(multi_index_incomplete.spatial_dimension == dim_incomplete)



"""
Module with functions required for converting different transformation formats into each other
and to construct a full size array for testing purposes (comparison with the regular conversion matrices).

.. todo::
    implement all other conversion functions
"""
from typing import no_type_check

import numpy as np
from numba import njit

from minterpy.global_settings import (
    ARRAY,
    FLOAT_DTYPE,
    INT_DTYPE,
    TRAFO_DICT,
    TYPED_LIST,
)


@no_type_check
@njit(cache=True)
def merge_trafo_dict(trafo_dict: TRAFO_DICT, leaf_positions: ARRAY) -> ARRAY:
    """Reconstructing the global transformation matrix from the precomputed transformation in dictionary format.

    :param trafo_dict: transformation dictionary
    :param leaf_positions: leaf positions
    :return: reconstructed transformation matrix

    """
    # TODO use the same merging fct everywhere in order to remove redundancies
    # TODO convert into piecewise format first, create fct for this,

    last_leaf_idx = len(leaf_positions) - 1
    last_leaf_pos = leaf_positions[last_leaf_idx]
    last_leaf_size = trafo_dict[last_leaf_idx, last_leaf_idx].shape[0]
    expected_size = last_leaf_pos + last_leaf_size
    combined_matrix = np.zeros((expected_size, expected_size), dtype=FLOAT_DTYPE)
    for (
        (leaf_idx_l, leaf_idx_r),
        matrix_piece,
    ) in trafo_dict.items():
        start_pos_in = leaf_positions[leaf_idx_l]
        start_pos_out = leaf_positions[leaf_idx_r]

        # NOTE: the size of the required slices of the coefficient vectors
        # are implicitly encoded in the size of each transformation array piece!
        size_out, size_in = matrix_piece.shape
        end_pos_in = start_pos_in + size_in
        end_pos_out = start_pos_out + size_out

        window = combined_matrix[start_pos_out:end_pos_out, start_pos_in:end_pos_in]
        window[:] = matrix_piece

    return combined_matrix


@njit(cache=True)
def factorised_2_piecewise(
    first_leaf_solution, leaf_factors, leaf_positions, leaf_sizes
):
    """Computes the actual matrix pieces of a transformation in factorised format explicitly.

    :param first_leaf_solution:
    :param leaf_factors:
    :param leaf_positions:
    :param leaf_sizes:
    :return:

    Notes
    -----
    Useful for merging all the pieces into a single matrix.
    """
    matrix_pieces = []
    start_positions_1 = []
    start_positions_2 = []
    nr_of_leaves = len(leaf_positions)
    for node_idx_1 in range(nr_of_leaves):
        # "lower triangular form"
        for node_idx_2 in range(node_idx_1, nr_of_leaves):
            corr_factor = leaf_factors[node_idx_2, node_idx_1]
            if corr_factor == 0.0:
                continue

            size_in = leaf_sizes[node_idx_1]
            size_out = leaf_sizes[node_idx_2]

            transformation_piece = (
                first_leaf_solution[:size_out, :size_in] * corr_factor
            )

            matrix_pieces.append(transformation_piece)

            start_pos_in = leaf_positions[node_idx_1]
            start_pos_out = leaf_positions[node_idx_2]
            start_positions_1.append(start_pos_in)
            start_positions_2.append(start_pos_out)

    start_positions_1 = np.array(start_positions_1, dtype=INT_DTYPE)
    start_positions_2 = np.array(start_positions_2, dtype=INT_DTYPE)

    return matrix_pieces, start_positions_1, start_positions_2


@njit(cache=True)
def merge_trafo_piecewise(
    matrix_pieces: TYPED_LIST, start_positions_in: ARRAY, start_positions_out: ARRAY
) -> ARRAY:
    """Reconstructing the global transformation matrix from the precomputed transformation in piecewise format.

    :param matrix_pieces:
    :param start_positions_in:
    :param start_positions_out:
    :return:

    Notes
    -----
    This is only used for testing the equality of the transformation matrices of both global and barycentric
    transformation.
    """

    # TODO allow to only create a slice of the total matrix
    # ATTENTION: the last entry must belong to the last leaf node
    # -> the ordering of the matrix pieces is not irrelevant!
    last_leaf_idx = len(matrix_pieces) - 1
    last_leaf_pos = start_positions_in[last_leaf_idx]
    last_leaf_size = matrix_pieces[last_leaf_idx].shape[0]
    expected_size = last_leaf_pos + last_leaf_size
    combined_matrix = np.zeros((expected_size, expected_size), dtype=FLOAT_DTYPE)
    for matrix_piece, start_pos_in, start_pos_out in zip(
        matrix_pieces, start_positions_in, start_positions_out
    ):
        # NOTE: the size of the required slices of the coefficient vectors
        # are implicitly encoded in the size of each transformation array piece!
        size_out, size_in = matrix_piece.shape
        end_pos_in = start_pos_in + size_in
        end_pos_out = start_pos_out + size_out

        window = combined_matrix[start_pos_out:end_pos_out, start_pos_in:end_pos_in]
        window[:] = matrix_piece

    return combined_matrix


@njit(cache=True)
def merge_trafo_factorised(
    first_leaf_solution: ARRAY,
    leaf_factors: ARRAY,
    leaf_positions: ARRAY,
    leaf_sizes: ARRAY,
) -> ARRAY:
    """Reconstructing the global transformation matrix from the precomputed transformation in factorised format.

    :param first_leaf_solution:
    :param leaf_factors:
    :param leaf_positions:
    :param leaf_sizes:
    :return:

    """
    trafo_piecewise = factorised_2_piecewise(
        first_leaf_solution, leaf_factors, leaf_positions, leaf_sizes
    )
    return merge_trafo_piecewise(*trafo_piecewise)



"""
set of package wide utility functions
"""
from typing import Union

import numpy as np

from minterpy.core.verification import (
    check_dtype,
    convert_eval_output,
    rectify_eval_input,
    rectify_query_points,
)
from minterpy.global_settings import DEBUG, FLOAT_DTYPE, INT_DTYPE
from minterpy.jit_compiled_utils import eval_newton_monomials_multiple


def lp_norm(arr, p, axis=None, keepdims: bool = False):
    """Robust lp-norm function.

    Works essentially like ``numpy.linalg.norm``, but is numerically stable for big arguments.

    :param arr: Input array.
    :type arr: np.ndarray

    :param axis: If axis is an integer, it specifies the axis of x along which to compute the vector norms. If axis is a 2-tuple, it specifies the axes that hold 2-D matrices, and the matrix norms of these matrices are computed. If axis is None then either a vector norm (when x is 1-D) or a matrix norm (when x is 2-D) is returned. The default is :class:`None`.
    :type axis: {None, int, 2-tuple of int}, optional

    :param keepdims: If this is set to True, the axes which are normed over are left in the result as dimensions with size one. With this option the result will broadcast correctly against the original ``arr``.
    :type keepdims: bool, optional
    """

    a = np.abs(arr).max()
    if a == 0.0:  # NOTE: avoid division by 0
        return 0.0
    return a * np.linalg.norm(arr / a, p, axis, keepdims)


def cartesian_product(*arrays: np.ndarray) -> np.ndarray:
    """
    Build the cartesian product of any number of 1D arrays.

    :param arrays: List of 1D array_like.
    :type arrays: list

    :return: Array of all combinations of elements of the input arrays (a cartesian product).
    :rtype: np.ndarray

    Examples
    --------
    >>> x = np.array([1,2,3])
    >>> y = np.array([4,5])
    >>> cartesian_product(x,y)
    array([[1, 4],
           [1, 5],
           [2, 4],
           [2, 5],
           [3, 4],
           [3, 5]])

    >>> s= np.array([0,1])
    >>> cartesian_product(s,s,s,s)
    array([[0, 0, 0, 0],
           [0, 0, 0, 1],
           [0, 0, 1, 0],
           [0, 0, 1, 1],
           [0, 1, 0, 0],
           [0, 1, 0, 1],
           [0, 1, 1, 0],
           [0, 1, 1, 1],
           [1, 0, 0, 0],
           [1, 0, 0, 1],
           [1, 0, 1, 0],
           [1, 0, 1, 1],
           [1, 1, 0, 0],
           [1, 1, 0, 1],
           [1, 1, 1, 0],
           [1, 1, 1, 1]])

    """
    la = len(arrays)
    dtype = np.result_type(*arrays)
    arr = np.empty([len(a) for a in arrays] + [la], dtype=dtype)
    for i, a in enumerate(np.ix_(*arrays)):
        arr[..., i] = a
    return arr.reshape(-1, la)


def lp_sum(arr: np.ndarray, p: Union[float, int]) -> Union[float, int]:
    """
    Sum of powers, i.e. lp-norm to the lp-degree.

    :param arr: 2D-array to be lp-summed
    :type arr: np.ndarray

    :param p: Power for each element in the lp-sum
    :type p: Real

    :return: lp-sum over the last axis of the input array powered by the given power
    :rtype: np.ndarray

    Notes
    -----
    - equivalent to ```lp_norm(arr,p,axis=-1)**p``` but more stable then the implementation using `np.linalg.norm` (see `numpy #5697`_  for more informations)

    .. _numpy #5697:
        https://github.com/numpy/numpy/issues/5697
    """
    return np.sum(np.power(arr, p), axis=-1)


def chebychev_2nd_order(n: int):  # 2nd order
    """Factory function of Chebychev points of the second kind.

    :param n: Degree of the point set, i.e. number of Chebychev points.
    :type n: int

    :return: Array of Chebychev points of the second kind.
    :rtype: np.ndarray

    .. todo::
        - rename this function
        - rename the parameter ``n``.
    """
    if n == 0:
        return np.zeros(1, dtype=FLOAT_DTYPE)
    if n == 1:
        return np.array([-1.0, 1.0], dtype=FLOAT_DTYPE)
    return np.cos(np.arange(n, dtype=FLOAT_DTYPE) * np.pi / (n - 1))


def gen_chebychev_2nd_order_leja_ordered(n: int):
    """Factory function of Leja ordered Chebychev points of the second kind.

    :param n: Degree of the point set, i.e. number of Chebychev points (plus one!).
    :type n: int

    :return: Array of Leja ordered Chebychev points of the second kind.
    :rtype: np.ndarray

    .. todo::
        - rename this function
        - rename the parameter ``n``.
        - refactor this function to remove all the loops.
        - make the arguments equivalent to ``chebychev_2nd_order``, i.e. if the degree ``n`` is passed, the number of points shall be ``n`` (not ``n+1``).
    """
    n = int(n)
    points1 = chebychev_2nd_order(n + 1)[::-1]
    points2 = points1  # TODO
    ord = np.arange(1, n + 1)

    lj = np.zeros([1, n + 1])
    lj[0] = 0
    m = 0

    for k in range(0, n):
        jj = 0
        for i in range(0, n - k):
            P = 1
            for j in range(k + 1):
                idx_pts = int(lj[0, j])
                P = P * (points1[idx_pts] - points1[ord[i]])
            P = np.abs(P)
            if P >= m:
                jj = i
                m = P
        m = 0
        lj[0, k + 1] = ord[jj]
        ord = np.delete(ord, jj)

    leja_points = np.zeros([n + 1, 1])
    for i in range(n + 1):
        leja_points[i, 0] = points2[int(lj[0, i])]
    return leja_points


def eval_newton_monomials(
    x: np.ndarray,
    exponents: np.ndarray,
    generating_points: np.ndarray,
    verify_input: bool = False,
    triangular: bool = False,
) -> np.ndarray:
    """Newton evaluation function.

    Compute the value of each Newton monomial on each given point. Internally it uses ``numba`` accelerated evaluation function.

    :param x: The points to evaluate the polynomials on.
    :type x: np.ndarray
    :param exponents: the multi indices "alpha" for every Newton polynomial corresponding to the exponents of this "monomial"
    :type exponents: np.ndarray
    :param generating_points: Nodes where the Newton polynomial lives on. (Or the points which generate these nodes?!)
    :type generating_points: np.ndarray
    :param verify_input: weather the data types of the input should be checked. Turned off by default for performance.
    :type verify_input: bool
    :param triangular: weather or not the output will be of lower triangular form. This will skip the evaluation of some values. Defaults to :class:`False`.
    :type triangular: bool

    :return: the value of each Newton polynomial on each point. The output shape is ``(k, N)``, where ``k`` is the number of points and ``N`` is the number of coeffitions of the Newton polyomial.
    :rtype: np.ndarray

    .. todo::
        - rename ``generation_points`` according to :class:`Grid`.
        - use instances of :class:`MultiIndex` and/or :class:`Grid` instead of the array representations of them.
        - ship this to the submodule ``newton_polynomials``.
        - Refactor the "triangular" parameter, the Newton monomials
          only becomes triangular if evaluated at the unisolvent nodes.
          So it needs a special function instead of parametrizing this function
          that can give a misleading result.

    See Also
    --------
    eval_all_newt_polys : concrete ``numba`` accelerated implementation of polynomial evaluation in Newton base.
    """
    N, m = exponents.shape
    nr_points, x = rectify_query_points(
        x, m
    )  # also working for 1D array, reshape into required shape
    if verify_input:
        check_dtype(x, FLOAT_DTYPE)
        check_dtype(exponents, INT_DTYPE)

    # NOTE: the downstream numba-accelerated function does not support kwargs,
    # so the maximum exponent per dimension must be computed here
    max_exponents = np.max(exponents, axis=0)

    # Create placeholders for the final and intermediate results
    result_placeholder = np.empty((nr_points, N), dtype=FLOAT_DTYPE)
    prod_placeholder = np.empty(
        (np.max(max_exponents) + 1, m), dtype=FLOAT_DTYPE
    )

    # Compute the Newton monomials on all the query points
    eval_newton_monomials_multiple(
        x,
        exponents,
        generating_points,
        max_exponents,
        prod_placeholder,
        result_placeholder,
        triangular,
    )

    return result_placeholder


def eval_newton_polynomials(
    xx: np.ndarray,
    coefficients: np.ndarray,
    exponents: np.ndarray,
    generating_points: np.ndarray,
    verify_input: bool = False,
    batch_size: int = None,
):
    """Evaluate the polynomial(s) in Newton form at multiple query points.

    Iterative implementation of polynomial evaluation in Newton form

    This version able to handle both:
        - list of input points x (2D input)
        - list of input coefficients (2D input)

    Here we use the notations:
        - ``n`` = polynomial degree
        - ``N`` = amount of coefficients
        - ``k`` = amount of points
        - ``p`` = amount of polynomials


    .. todo::
        - idea for improvement: make use of the sparsity of the exponent matrix and avoid iterating over the zero entries!
        - refac the explanation and documentation of this function.
        - use instances of :class:`MultiIndex` and/or :class:`Grid` instead of the array representations of them.
        - ship this to the submodule ``newton_polynomials``.

    :param xx: Arguemnt array with shape ``(m, k)`` the ``k`` points to evaluate on with dimensionality ``m``.
    :type xx: np.ndarray
    :param coefficients: The coefficients of the Newton polynomials.
        NOTE: format fixed such that 'lagrange2newton' conversion matrices can be passed
        as the Newton coefficients of all Lagrange monomials of a polynomial without prior transponation
    :type coefficients: np.ndarray, shape = (N, p)
    :param exponents: a multi index ``alpha`` for every Newton polynomial corresponding to the exponents of this ``monomial``
    :type exponents: np.ndarray, shape = (m, N)
    :param generating_points: Grid values for every dimension (e.g. Leja ordered Chebychev values).
        the values determining the locations of the hyperplanes of the interpolation grid.
        the ordering of the values determine the spacial distribution of interpolation nodes.
        (relevant for the approximation properties and the numerical stability).
    :type generating_points: np.ndarray, shape = (m, n+1)
    :param verify_input: weather the data types of the input should be checked. turned off by default for speed.
    :type verify_input: bool, optional
    :param batch_size: batch size of query points
    :type batch_size: int, optional

    :raise TypeError: If the input ``generating_points`` do not have ``dtype = float``.

    :return: (k, p) the value of each input polynomial at each point. TODO squeezed into the expected shape (1D if possible). Notice, format fixed such that the regression can use the result as transformation matrix without transponation

    Notes
    -----
    - This method is faster than the recursive implementation of ``tree.eval_lp(...)`` for a single point and a single polynomial (1 set of coeffs):
        - time complexity: :math:`O(mn+mN) = O(m(n+N)) = ...`
        - pre-computations: :math:`O(mn)`
        - evaluation: :math:`O(mN)`
        - space complexity: :math:`O(mn)` (precomputing and storing the products)
        - evaluation: :math:`O(0)`
    - advantage:
        - just operating on numpy arrays, can be just-in-time (jit) compiled
        - can evaluate multiple polynomials without recomputing all intermediary results

    See Also
    --------
    evaluate_multiple : ``numba`` accelerated implementation which is called internally by this function.
    convert_eval_output: ``numba`` accelerated implementation of the output converter.
    """

    # Rectify the inputs
    # TODO: Refactor this rectify function
    verify_input = verify_input or DEBUG
    _, coefficients, _, n_points, _, xx = \
        rectify_eval_input(xx, coefficients, exponents, verify_input)

    # Get batch size
    # TODO: Verify the batch size
    if batch_size is None or batch_size >= n_points:
        newton_monomials = eval_newton_monomials(
            xx,
            exponents,
            generating_points,
            verify_input,
            False
        )
        results = newton_monomials @ coefficients
    else:
        # Evaluate the Newton polynomials in batches
        results = eval_newton_polynomials_batch(
            xx,
            coefficients,
            exponents,
            generating_points,
            batch_size
        )

    return convert_eval_output(results)


def eval_newton_polynomials_batch(
    xx: np.ndarray,
    coefficients: np.ndarray,
    exponents: np.ndarray,
    generating_points: np.ndarray,
    batch_size: int
):
    """Evaluate the polynomial in Newton form in batches of query points.

    Notes
    -----
    - It is assumed that the inputs have all been verified and rectified.
    - It would be more expensive to evaluate smaller batch sizes
      but with a less smaller memory footprint in any given iteration.
    - If memory does not permit whole evaluation of query points,
      consider using smaller but not the smallest batch size (i.e., not 1).
    """

    # Get some important numbers
    n_points = xx.shape[0]
    n_polynomials = coefficients.shape[1]

    # Create a placeholder for the results
    results_placeholder = np.empty(
        (n_points, n_polynomials), dtype=FLOAT_DTYPE
    )

    # Batch processing: evaluate the polynomials at a batch of query points
    n_batches = n_points // batch_size + 1
    for idx in range(n_batches):
        start_idx = idx * batch_size

        if idx == n_batches - 1:
            end_idx = idx * batch_size + n_points % batch_size
        else:
            end_idx = (idx + 1) * batch_size

        # Get the current batch of query points
        xx_batch = xx[start_idx:end_idx, :]

        # Compute the Newton monomials for the batch
        newton_monomials = eval_newton_monomials(
            xx_batch,
            exponents,
            generating_points,
            False,
            False
        )

        # Compute the polynomial values for the batch
        results_placeholder[start_idx:end_idx, :] = \
            newton_monomials @ coefficients

    return results_placeholder



"""
Concrete implementations for Operator classes.

The following implementations are provided:

- MatrixOperator
- BarycentricDictOperator
- BarycentricFactorisedOperator
- BarycentricPiecewiseOperator

"""

from abc import abstractmethod
from typing import Optional, Union

import numpy as np
from _warnings import warn

from minterpy.core.ABC import OperatorABC
from minterpy.global_settings import ARRAY, FLOAT_DTYPE

from ..matrix_operator import MatrixOperator
from .conversion import merge_trafo_dict, merge_trafo_factorised, merge_trafo_piecewise
from .transformation_fcts import (
    transform_barycentric_dict,
    transform_barycentric_factorised,
    transform_barycentric_piecewise,
)


class BarycentricOperator(OperatorABC):
    """Base class for the barycentric transformation operators.
       Specialized versions (dict, factorised, and piecewise) are derived from this base class.

    Attributes
    ----------
    array_representation

    """

    _array_representation: Optional[ARRAY] = None

    @property
    def array_representation(self) -> ARRAY:
        """Reconstructs the global transformation matrix."""
        if self._array_representation is None:
            warn(
                "building a full transformation matrix from a barycentric transformation. this is inefficient."
            )
            # NOTE: 'self' arg must not be passed to the the merging fcts (@staticmethod)
            full_array = self.__class__.merging_fct(*self.transformation_data)  # type: ignore
            self._array_representation = full_array
        return self._array_representation

    @staticmethod
    @abstractmethod
    def transformation_fct(coeffs_in, coeffs_out_placeholder, *args):
        """Abstract method for executing the potentially decomposed linear transformation."""
        pass

    @staticmethod
    @abstractmethod
    def merging_fct(*args):
        """Abstract method for reconstructing the global matrix from the decomposition."""
        pass

    def __matmul__(
        self, other: Union[OperatorABC, ARRAY]
    ) -> Union[MatrixOperator, ARRAY]:
        """Applies the transformation operator on the input."""
        if isinstance(other, OperatorABC):
            # the input is another transformation
            # workaround: return matrix operator constructed from the matrix product
            # TODO support this natively
            # TODO which transformation object should be passed?
            return MatrixOperator(
                self.transformation, self.array_repr_full @ other.array_repr_full
            )

        # TODO support "separate multi index" transformations

        # assuming the input are coefficients which should be transformed
        coeffs_in = other  # alias
        # use an output placeholder (for an increases compatibility with Numba JIT compilation)
        # initialise the placeholder with 0
        coeffs_out_placeholder = np.zeros(coeffs_in.shape, dtype=FLOAT_DTYPE)
        # trafo_dict, leaf_positions = self.data_container

        # NOTE: 'self' arg must not be passed to the the transformation fcts (@staticmethod)
        self.__class__.transformation_fct(
            coeffs_in, coeffs_out_placeholder, *self.transformation_data
        )  # type: ignore
        return coeffs_out_placeholder

    def _get_array_repr(self):
        """Reconstructs the global transformation matrix."""
        return self.array_representation


class BarycentricDictOperator(BarycentricOperator):
    """Concrete implementation of the BarycentricOperator given by the edge case given by decomposition to the 1D
    atomic sub-problems.
    """

    transformation_fct = transform_barycentric_dict
    merging_fct = merge_trafo_dict


class BarycentricFactorisedOperator(BarycentricOperator):
    """Concrete implementation of the BarycentricOperator given by the edge case given by realizing the factorised
    copied of the basic 1D atomic sub-problem.
    """

    transformation_fct = transform_barycentric_factorised
    merging_fct = merge_trafo_factorised


class BarycentricPiecewiseOperator(BarycentricOperator):
    """Concrete implementation of the BarycentricOperator used to make comparisons with the global matrix possible by
    zooming into corresponding submatrix.
    """

    transformation_fct = transform_barycentric_piecewise
    merging_fct = merge_trafo_piecewise



"""
This module contains the concrete implementation of ordinary regression.
"""
import numpy as np
import scipy
from typing import Optional, Type, Union, Tuple, Callable

from minterpy.transformations import LagrangeToNewton
from minterpy.utils import eval_newton_monomials, eval_newton_polynomials
from minterpy.core.ABC import MultivariatePolynomialSingleABC
from minterpy.polynomials import (
    LagrangePolynomial, NewtonPolynomial, CanonicalPolynomial
)
from minterpy.core.multi_index import MultiIndexSet
from minterpy.core.grid import Grid

from .regression_abc import RegressionABC

__all__ = ["OrdinaryRegression"]


TOriginPoly = Union[
    Type[LagrangePolynomial], Type[NewtonPolynomial], Type[CanonicalPolynomial]
]


class OrdinaryRegression(RegressionABC):
    """Implementation of an ordinary (weighted/unweighted) poly. regression.

    Parameters
    ----------
    multi_index : MultiIndexSet, optional
        The multi-index set that defines the underlying polynomial.
        This parameter is optional if a grid is specified; in that case,
        the multi-index set is set to the one attached to the grid.
    grid : Grid, optional
        The grid where the polynomial lives.
        This parameter is optional if a multi-index set is specified;
        in that case, the grid is constructed from the specified
        multi-index set.
    origin_poly : TOriginPoly, optional
        The polynomial basis on which the regression is carried out.
        This parameter is optional and, by default, is set to the Lagrange
        polynomial.
    """

    def __init__(
            self,
            multi_index: Optional[MultiIndexSet] = None,
            grid: Optional[Grid] = None,
            origin_poly: TOriginPoly = LagrangePolynomial,
    ):
        if multi_index is None and grid is None:
            raise ValueError(
                "Either multi-index set or grid must be specified!"
            )

        # Initialize and verify the grid
        if grid is None:
            # Create a grid from a verified multi-index set
            _verify_multi_index(multi_index)
            self._grid = Grid(multi_index)
        else:
            _verify_grid(grid)
            self._grid = grid

        # Verify multi-index set
        if multi_index is None:
            self._multi_index = grid.multi_index
        else:
            # Verify the multi-index set and its relation with the grid
            _verify_multi_index(multi_index)
            _verify_grid(self.grid, multi_index)
            self._multi_index = multi_index

        # Initialize the origin polynomial basis
        self._origin_poly = origin_poly(
            multi_index=self._multi_index, grid=self._grid
        )

        # Initialize additional internal attributes
        self._regression_matrix = None
        self._coeffs = None
        self._loocv_error = None
        self._regfit_l1_error = None
        self._regfit_l2_error = None
        self._eval_poly = None

    @property
    def multi_index(self):
        return self._multi_index

    @property
    def loocv_error(self):
        return self._loocv_error

    @property
    def regfit_l1_error(self):
        return self._regfit_l1_error

    @property
    def regfit_l2_error(self):
        return self._regfit_l2_error

    @property
    def grid(self):
        return self._grid

    @property
    def origin_poly(self):
        return self._origin_poly

    @property
    def eval_poly(self):
        return self._eval_poly

    def get_regression_matrix(self, xx):
        """Get the regression matrix on a set of evaluation points."""
        return compute_regression_matrix(self._origin_poly, xx)

    def fit(
        self,
        xx: np.ndarray,
        yy: np.ndarray,
        weights: np.ndarray = None,
        lstsq_solver: Union[str, Callable] = "lstsq",
        **kwargs,
    ):
        """Fit the ordinary polynomial regression model.

        Parameters
        ----------
        xx : np.ndarray
            Input matrix, also known as the training inputs.
        yy : np.ndarray
            Response vector, observed or evaluated at ``xx``.
        weights: np.ndarray, optional
            Individual weights for each input points.
            The default is ``None``.
        lstsq_solver : str, optional
            Least-square solver. The default is ``lstsq`` from SciPy.

        Returns
        -------
        None
            The instance itself is updated with a fitted polynomial.
            After a successful fitting, the instance can be evaluated on
            a set of query points.

        Notes
        -----
        - ``**kwargs`` may take additional keyword arguments that are passed
          to the selected least-square solver. Refer to the documentation
          of the solver for the list of supported keyword arguments.
        """
        # Get the regression matrix on the evaluation points
        self._regression_matrix = self.get_regression_matrix(xx)

        # Solve the least-squares problem to obtain the coefficients
        self._coeffs = solve_least_squares(
            self._regression_matrix, yy, weights, lstsq_solver, **kwargs
        )

        # Compute some error metrics
        self._regfit_l1_error = compute_regfit_l1_error(
            self._regression_matrix, self._coeffs, yy
        )
        self._regfit_l2_error = compute_regfit_l2_error(
            self._regression_matrix, self._coeffs, yy
        )
        self._loocv_error = compute_loocv_error(
            self._regression_matrix, self._coeffs, yy, weights
        )

        # Create a polynomial that can be evaluated
        if isinstance(self.origin_poly, LagrangePolynomial):
            # Lagrange can't be evaluated, transform to Newton first
            lag_poly = LagrangePolynomial(
                multi_index=self.multi_index,
                coeffs=self._coeffs,
                grid=self.grid
            )
            l2n = LagrangeToNewton(lag_poly)
            eval_poly = l2n()
        else:
            # Other polynomials can be directly evaluated
            eval_poly = type(self.origin_poly)(
                multi_index=self.multi_index,
                coeffs=self._coeffs,
                grid=self.grid
            )

        self._eval_poly = eval_poly

    def show(self):
        output = f"Ordinary Polynomial Regression\n" \
                 f"------------------------------\n" \
                 f"Spatial dimension: {self.multi_index.spatial_dimension}\n" \
                 f"Poly. degree     : {self.multi_index.poly_degree}\n"\
                 f"Lp-degree        : {self.multi_index.lp_degree}\n" \
                 f"Origin poly.     : {type(self.origin_poly)}\n"
        if self.regfit_l1_error is not None:
            output += f"{'Error':<19s}  {'Absolute':>10s}  {'Relative':>10s}\n"\
                      f"L1 Regression Fit   {self.regfit_l1_error[0]:10.5e} " \
                      f"{self.regfit_l1_error[1]:10.5e}\n" \
                      f"L2 Regression Fit   {self.regfit_l2_error[0]:10.5e} " \
                      f"{self.regfit_l2_error[1]:10.5e}\n" \
                      f"LOO CV              {self.loocv_error[0]:10.5e} " \
                      f"{self.loocv_error[1]:10.5e}"

        print(output)

    def predict(self, xx) -> np.ndarray:
        if self._eval_poly is None:
            raise TypeError("Ordinary regression model is not fitted, "
                            "regression polynomial can't be evaluated!")
        else:
            return self._eval_poly(xx)


def compute_regression_matrix(
    basis_poly: MultivariatePolynomialSingleABC,
    xx: np.ndarray
) -> np.ndarray:
    """Construct a regression matrix of the chosen polynomial basis.

    The regression matrix is constructed by evaluation the monomials of
    the chosen basis on all the input points.

    TODO: This is a temporary solution as ideally each polynomial basis
          can call its own "eval_monomials_on" method to construct
          the regression matrix on the training points.
    """
    exponents = basis_poly.multi_index.exponents
    generating_points = basis_poly.grid.generating_points

    regression_matrix = None
    # Evaluate the basis at the trainining points
    # TODO: should have a method basis_poly.eval_monomials(xx),
    #       below are temporary solutions.
    if isinstance(basis_poly, LagrangePolynomial):
        # Compute the coefficients of the Newton basis
        # for the Lagrange monomials
        newton_coeffs = LagrangeToNewton(
            basis_poly
        ).transformation_operator.array_repr_full

        # Evaluate the Newton polynomials representing the Lagrange monomials
        # on the input points
        regression_matrix = eval_newton_polynomials(
            xx, newton_coeffs, exponents, generating_points
        )

        # Edge case: only a single training point is given
        if xx.shape[0] == 1:
            regression_matrix = regression_matrix[np.newaxis, :]

        # Edge case: only a single Lagrange basis is given
        if regression_matrix.ndim == 1:
            regression_matrix = regression_matrix[:, np.newaxis]

    elif isinstance(basis_poly, NewtonPolynomial):
        regression_matrix = eval_newton_monomials(
            xx, exponents, generating_points
        )

    elif isinstance(basis_poly, CanonicalPolynomial):
        regression_matrix = np.prod(
            np.power(xx[:, None, :], exponents[None, :, :]), axis=-1
        )

    else:
        raise TypeError(f"Polynomial {type(basis_poly)} is not supported!")

    return regression_matrix


def solve_least_squares(
    regression_matrix: np.ndarray,
    yy: np.ndarray,
    weights: np.ndarray,
    lstsq_solver: Union[str, Callable],
    **kwargs,
) -> np.ndarray:
    """Solve the least-squares problem with a specified solver.

    The following linear solvers are available as pre-defined:

    - ``lstsq``: least-squares solver (SciPy)
    - ``inv``: (Gram) matrix inversion (NumPy)
    - ``pinv``: pseudo-inversion of the regression matrix (NumPy)
    - ``dgesv``: LU with full pivoting solver (SciPy)
    - ``dsysv``: Diagonal pivoting solver (SciPy)
    - ``dposv``: Cholesky decomposition solver (SciPy)
    - ``qr``: QR-decomposition-based solver (NumPy)
    - ``svd``: SVD-based solver (NumPy)

    Parameters
    ----------
    regression_matrix : np.ndarray
        Regression matrix constructed on the chosen polynomial basis and
        training points.
    yy : np.ndarray
        Function values that correspond to the training points.
    weights: np.ndarray
        The weights associated with the function values.
    lstsq_solver : Union[str, Callable]
        Chosen least-square solvers, may be chosen from a pre-defined solver
        using a string or a user-defined function that satisfies a common
        interface.

    Returns
    -------
    np.ndarray
        The coefficients of the regression polynomial.
    """
    # If the weights are given, make a diagonal matrix out of it
    if weights is not None:
        if weights.ndim == 1:
            weights = np.diag(weights)

    if isinstance(lstsq_solver, Callable):
        # Solve with a user-defined least-squares solver
        coeffs = lstsq_solver(regression_matrix, yy, weights, **kwargs)

        return coeffs

    lstsq_solver = lstsq_solver.lower()

    if lstsq_solver == "lstsq":
        # Least-squares solver from SciPy (based on SVD)
        coeffs = _solve_lstsq(regression_matrix, yy, weights, **kwargs)

    elif lstsq_solver == "inv":
        # Matrix inversion from NumPy (inversion of the Gram matrix)
        coeffs = _solve_inv(regression_matrix, yy, weights)

    elif lstsq_solver == "pinv":
        # Pseudo-inverse (of the regression matrix) from NumPy
        coeffs = _solve_pinv(regression_matrix, yy, weights, **kwargs)

    elif lstsq_solver == "dgesv":
        # LU w/ full pivoting solver from SciPy
        coeffs = _solve_dgesv(regression_matrix, yy, weights, **kwargs)

    elif lstsq_solver == "dsysv":
        # Diagonal pivoting solver from SciPy
        coeffs = _solve_dsysv(regression_matrix, yy, weights, **kwargs)

    elif lstsq_solver == "dposv":
        # Cholesky decomposition solver from SciPy
        coeffs = _solve_dposv(regression_matrix, yy, weights, **kwargs)

    elif lstsq_solver == "qr":
        # QR-decomposition-based solver from NumPy
        coeffs = _solve_qr(regression_matrix, yy, weights)

    elif lstsq_solver == "svd":
        # SVD-based solver from NumPy
        coeffs = _solve_svd(regression_matrix, yy, weights)

    else:
        raise NotImplementedError(f"{lstsq_solver} solver not implemented!")

    return coeffs


def compute_loocv_error(
    regression_matrix: np.ndarray,
    coeffs: np.ndarray,
    yy: np.ndarray,
    weights: np.ndarray
) -> Tuple[float, float]:
    """Calculate the leave-one-out (LOO) cross-validation (CV) errors.

    Parameters
    ----------
    regression_matrix : np.ndarray
        Regression matrix constructed on the chosen polynomial basis and
        training points.
    coeffs : np.ndarray
        Coefficients of the regression polynomial.
    yy : np.ndarray
        Function values that correspond to the training points.
    weights: np.ndarray
        Weights associated with the function values.

    Returns
    -------
    Tuple[float, float]
        The leave-one-out cross-validation errors, in absolute
        and relative (normalized) terms. The normalization is with respect to
        the function values sample variance.
    """
    # If the weights are given, make a diagonal matrix out of it
    if weights is None:
        weights = np.eye(len(yy))
    else:
        if weights.ndim == 1:
            weights = np.diag(weights)

    if regression_matrix.shape[0] >= regression_matrix.shape[1]:
        # Determined or over-determined system
        hi = np.diag(
            regression_matrix @ np.linalg.pinv(
                regression_matrix.T @ weights @ regression_matrix
            ) @ regression_matrix.T @ weights
        )
        loo_cv = ((yy - regression_matrix @ coeffs) / (1 - hi))**2
        idx = np.where(np.isnan(loo_cv))
        loo_cv[idx] = np.infty
        loo_cv_error = np.mean(loo_cv)
    else:
        # NOTE: Under-determined system, unreliable results
        loo_cv_error = np.infty

    return loo_cv_error, loo_cv_error / np.var(yy)


def compute_regfit_l1_error(
    regression_matrix: np.ndarray,
    coeffs: np.ndarray,
    yy: np.ndarray,
) -> Tuple[float, float]:
    """Calculate the absolute and normalized L1 regression fit error.

    Calculate the absolute and normalized L2 regression fit error.

    Parameters
    ----------
    regression_matrix : np.ndarray
        Regression matrix constructed on the chosen polynomial basis and
        training points.
    coeffs : np.ndarray
        Coefficients of the regression polynomial.
    yy : np.ndarray
        Function values that correspond to the training points.
    weights: np.ndarray
        Weights associated with the function values.

    Returns
    -------
    Tuple[float, float]
        The L1 regression fit error (i.e., max. of absolute error), in absolute
        and relative (normalized) terms. The normalization is with respect to
        the function values sample standard deviation.
    """
    l1_error = np.max(np.abs(yy - regression_matrix @ coeffs))

    return l1_error, l1_error / np.std(yy)


def compute_regfit_l2_error(
    regression_matrix: np.ndarray,
    coeffs: np.ndarray,
    yy: np.ndarray,
) -> Tuple[float, float]:
    """Calculate the absolute and normalized L2 regression fit error.

    Parameters
    ----------
    regression_matrix : np.ndarray
        Regression matrix constructed on the chosen polynomial basis and
        training points.
    coeffs : np.ndarray
        Coefficients of the regression polynomial.
    yy : np.ndarray
        Function values that correspond to the training points.
    weights: np.ndarray
        Weights associated with the function values.

    Returns
    -------
    Tuple[float, float]
        The L2 regression fit error (i.e., mean-squared error), in absolute
        and relative (normalized) terms. The normalization is with respect to
        the function values sample variance.
    """
    l2_error = float(np.mean((yy - regression_matrix @ coeffs) ** 2))

    return l2_error, l2_error / np.var(yy)


def _verify_multi_index(multi_index: MultiIndexSet):
    """Verify the instance of MultiIndexSet passed to the constructor."""
    if not isinstance(multi_index, MultiIndexSet):
        raise TypeError(f"Unexpected type {type(multi_index)} "
                        f"of the input multi-index set!")


def _verify_grid(grid: Grid, multi_index: Optional[MultiIndexSet] = None):
    """Verify the instance of Grid passed to the constructor"""
    if not isinstance(grid, Grid):
        raise TypeError(
            f"Unexpected type {type(grid)} of the input grid!"
        )

    if multi_index is not None:
        if not grid.multi_index.is_super_index_set_of(multi_index):
            raise ValueError(
                "The multi-indices of a polynomial must be a subset of "
                "the multi-indices of the grid in use!"
            )


def _solve_lstsq(
    rr: np.ndarray,
    yy: np.ndarray,
    ww: np.ndarray,
    **kwargs,
) -> np.ndarray:
    """Solve a least-squares problem using lstsq solver (SciPy)."""
    if ww is None:
        coeffs, _, _, _ = scipy.linalg.lstsq(rr, yy, **kwargs)
    else:
        coeffs, _, _, _ = scipy.linalg.lstsq(
            rr.T @ ww @ rr, rr.T @ ww @ yy, **kwargs
        )

    return coeffs


def _solve_inv(
    rr: np.ndarray,
    yy: np.ndarray,
    ww: np.ndarray,
) -> np.ndarray:
    """Solve a system of linear equations via matrix inversion."""
    if ww is None:
        ww = np.eye(len(yy))

    coeffs = np.linalg.inv(rr.T @ ww @ rr) @ rr.T @ ww @ yy

    return coeffs


def _solve_pinv(
    rr: np.ndarray,
    yy: np.ndarray,
    ww: np.ndarray,
    **kwargs,
) -> np.ndarray:
    """Solve a least-squares problem via pseudo-inverse of the regr. matrix."""
    if ww is None:
        coeffs = np.linalg.pinv(rr, **kwargs) @ yy
    else:
        coeffs = np.linalg.pinv(rr.T @ ww @ rr, **kwargs) @ rr.T @ ww @ yy

    return coeffs


def _solve_dgesv(
    rr: np.ndarray,
    yy: np.ndarray,
    ww: np.ndarray,
    **kwargs,
) -> np.ndarray:
    """Solve a least-squares problem using LU w/ full pivoting solver."""
    if ww is None:
        ww = np.eye(len(yy))

    _, _, coeffs, _ = scipy.linalg.lapack.dgesv(
        rr.T @ ww @ rr, rr.T @ ww @ yy, **kwargs
    )

    return coeffs


def _solve_dsysv(
    rr: np.ndarray,
    yy: np.ndarray,
    ww: np.ndarray,
    **kwargs,
) -> np.ndarray:
    """Solve a least-squares problem using a diagonal pivoting solver."""
    if ww is None:
        ww = np.eye(len(yy))

    _, _, coeffs, _ = scipy.linalg.lapack.dsysv(
            rr.T @ ww @ rr, rr.T @ ww @ yy, **kwargs
    )

    return coeffs


def _solve_dposv(
    rr: np.ndarray,
    yy: np.ndarray,
    ww: np.ndarray,
    **kwargs,
) -> np.ndarray:
    """Solve a least-squares problem using a Cholesky decompos. solver."""
    if ww is None:
        ww = np.eye(len(yy))

    _, coeffs, _ = scipy.linalg.lapack.dposv(
        rr.T @ ww @ rr, rr.T @ ww @ yy, **kwargs
    )

    return coeffs


def _solve_qr(
    rr: np.ndarray,
    yy: np.ndarray,
    ww: np.ndarray,
) -> np.ndarray:
    """Solve a least-squares problem using a QR-decomposition-based solver."""
    if ww is None:
        if rr.shape[0] >= rr.shape[1]:
            # Determined or over-determined system
            qq_star, rr_star = np.linalg.qr(rr)
            coeffs = np.linalg.solve(rr_star, qq_star.T @ yy)
        else:
            # Under-determined system
            qq_star, rr_star = np.linalg.qr(rr.T)
            coeffs = qq_star @ np.linalg.inv(rr_star.T) @ yy
    else:
        qq_star, rr_star = np.linalg.qr(rr.T @ ww @ rr)
        coeffs = scipy.linalg.solve_triangular(
            rr_star, qq_star.T @ rr.T @ ww @ yy
        )

    return coeffs


def _solve_svd(
    rr: np.ndarray,
    yy: np.ndarray,
    ww: np.ndarray,
) -> np.ndarray:
    """Solve a least-squares problem using an SVD-based solver."""
    if ww is None:
        uu, s, vv_t = np.linalg.svd(rr)

        # Compute the pseudo-inverse of the singular value matrix
        ss_inv = np.zeros((rr.shape[1], rr.shape[0]))
        np.fill_diagonal(ss_inv, 1 / s)

        coeffs = (vv_t.T @ ss_inv @ uu.T) @ yy
    else:
        rr_star = rr.T @ ww @ rr
        # SVD-based solver from NumPy
        uu, s, vv_t = np.linalg.svd(rr_star)

        # Compute the pseudo-inverse of the singular value matrix
        ss_inv = np.zeros((rr_star.shape[1], rr_star.shape[0]))
        np.fill_diagonal(ss_inv, 1 / s)

        yy_star = rr.T @ ww @ yy
        coeffs = (vv_t.T @ ss_inv @ uu.T) @ yy_star

    return coeffs


"""
Base class for polynomials in the canonical base.

"""
from copy import deepcopy

import numpy as np
from scipy.special import factorial

from minterpy.global_settings import DEBUG, FLOAT_DTYPE, INT_DTYPE
from minterpy.jit_compiled_utils import can_eval_mult, all_indices_are_contained

from ..core import MultiIndexSet
from ..core.ABC import MultivariatePolynomialSingleABC
from ..core.verification import convert_eval_output, rectify_eval_input, verify_domain
from ..core.utils import find_match_between

__all__ = ["CanonicalPolynomial"]


def dummy():
    """Placeholder function.

    .. warning::
      This function is not implemented yet!
    """
    raise NotImplementedError("This feature is not implemented yet.")


# Arithmetics


def _generic_canonical_add(mi1, c1, mi2, c2):
    """
    Generic add function based on exponents and coeffs (passed as arrays).

    Parameters
    ----------
    mi1 : array_like (shape = (mi1_len,dim))
        exponents of the first multi_index set.
    c1 : array_like (shape = (mi1_len,))
        coeffs related to mi1
    mi2 : array_like (shape = (mi2_len,dim))
        exponents of the second multi_index set.
    c2 : array_like (shape = (mi2_len,))
        coeffs related to mi2

    Returns
    -------
    res_mi : array_like
        Resulting exponents (essentially the union of mi1 and mi2) in lexicographical order.
    res_coeffs : array_like
        Resulting coefficients related to res_mi
        (essentially the sum of c1 and c2 where the exponents in mi1 and mi2 are the same and a concatenation of the rest.)

    Notes
    -----
    - there is no check if the shapes of the passed arrays match. This lies in the actual canonical_add function.
    """

    len1, dim1 = mi1.shape
    (
        len2,
        dim2,
    ) = (
        mi2.shape
    )  # assume m1 and m2 are same dimension -> expand_dim of the mi with smaller dim

    mi1r = mi1.reshape(len1, 1, dim1)
    mi2r = mi2.reshape(1, len2, dim1)

    add_cond = np.equal(mi1r, mi2r).all(
        axis=2, keepdims=False
    )  # where are mi1 and mi2 equal along the dim?
    add_cond1 = add_cond.any(axis=1)  # added condition for m1,c1
    not_add_cond1 = np.logical_not(add_cond1)  # not added condition for m1,c1
    add_cond2 = add_cond.any(axis=0)  # add condition for m2,c2
    not_add_cond2 = np.logical_not(add_cond2)  # not added condition for m2,c2

    # build resulting mi
    added_mi = mi1[add_cond1, :]  # shall be the same as mi2[add_cond2,:]
    not_added_mi = np.concatenate(
        (mi1[not_add_cond1, :], mi2[not_add_cond2, :]), axis=0
    )  # collect the rest
    res_mi = np.concatenate(
        (added_mi, not_added_mi), axis=0
    )  # summed mi shall be the first
    sort_args = np.lexsort(res_mi.T, axis=-1)

    # build resulting coeffs
    added_c = c1[add_cond1] + c2[add_cond2]
    not_added_c = np.concatenate((c1[not_add_cond1], c2[not_add_cond2]))
    res_c = np.concatenate((added_c, not_added_c))

    return res_mi[sort_args], res_c[sort_args]


def _match_dims(poly1, poly2, copy=None):
    """Dimensional expansion of two polynomial in order to match their spatial_dimensions.

    Parameters
    ----------
    poly1 : CanonicalPolynomial
        First polynomial in canonical basis
    poly2 : CanonicalPolynomial
        Second polynomial in canonical basis
    copy : bool
        If True, work on deepcopies of the passed polynomials (doesn't change the input).
        If False, inplace expansion of the passed polynomials

    Returns
    -------
    (poly1,poly2) : (CanonicalPolynomial,CanonicalPolynomial)
        Dimensionally matched polynomials in the same order as input.

    Notes
    -----
    - Maybe move this to the MultivariatePolynomialSingleABC since it shall be avialable for all poly bases
    """
    if copy is None:
        copy = True

    if copy:
        p1 = deepcopy(poly1)
        p2 = deepcopy(poly2)
    else:
        p1 = poly1
        p2 = poly2

    dim1 = p1.multi_index.spatial_dimension
    dim2 = p2.multi_index.spatial_dimension
    if dim1 >= dim2:
        p2.expand_dim(dim1)
    else:
        p1.expand_dim(dim2)
    return p1, p2


def _matching_internal_domain(poly1, poly2, tol=None):
    """Test if two polynomial have the same internal_domain.

    Works on polynomial with same spatial_dimension.

    Parameters
    ----------
    poly1, poly2 : CanonicalPolynomial
            Polynomials to check
    tol : float
            tolerance for matching floats in the domains. Default tol = 1e-16

    Returns
    -------
    match : bool
            True if both
    """
    if tol is None:
        tol = 1e-16
    return (
        np.less_equal(np.abs(poly1.internal_domain - poly2.internal_domain), tol)
        & np.less_equal(np.abs(poly1.user_domain - poly2.user_domain), tol)
    ).all()


def _canonical_add(poly1, poly2):
    """
    Addition of two polynomials in canonical basis.

    Parameters
    ----------
    poly1 : CanonicalPolynomial
        First summand of the sum
    poly2 : CanonicalPolynomial
        Second summand of the sum

    Returns
    -------
    summed_polynomial : CanonicalPolynomial
        The sum of the passed polynomials in canonical base

    Notes
    -----
     - This works only for the same domains!
     - This function will be called on `self,other` if `__add__` is called.
    """
    p1, p2 = _match_dims(poly1, poly2)
    # print(p1.internal_domain,p2.internal_domain) # here is the error!!!!
    if _matching_internal_domain(p1, p2):
        res_mi_arr, res_c = _generic_canonical_add(
            p1.multi_index.exponents, p1.coeffs, p2.multi_index.exponents, p2.coeffs
        )

        # warning: this is not general. we are *assuming* that the resultant lp_degree is
        # max of the two. This is done in order to avoid issue #37.
        max_lp_degree = np.max([p1.multi_index.lp_degree, p2.multi_index.lp_degree])
        res_mi = MultiIndexSet(res_mi_arr, lp_degree=max_lp_degree)
        return CanonicalPolynomial(
            res_mi,
            res_c,
            internal_domain=p1.internal_domain,
            user_domain=p1.user_domain,
        )
    else:
        raise NotImplementedError(
            "Addition is not implemented for canonical polynomials with different domains"
        )


def _canonical_sub(poly1, poly2):
    """
    Substraction of two polynomials in canonical basis.

    Parameters
    ----------
    poly1 : CanonicalPolynomial
        Minuend of the difference.
    poly2 : CanonicalPolynomial
        Subtrahend of the difference.

    Returns
    -------
    difference_polynomial : CanonicalPolynomial
        The difference of the passed polynomials in canonical base.

    Notes
    -----
    This works only for the same domains!
    """
    return _canonical_add(poly1, -poly2)


def _canonical_eval(pts: np.ndarray,exponents: np.ndarray, coeffs: np.ndarray):
    """
    Unsafe version of naive canonical evaluation

    :param pts: List of points, the polynomial must be evaluated on. Assumed shape: `(number_of_points,spatial_dimension)`.
    :type pts: np.ndarray

    :param exponents: Exponents from a multi-index set. Assumed shape:` (number_of_monomials, spatial_dimension)`.
    :type exponents: np.ndarray

    :param coeffs: List of coefficients. Assumed shape: `(number_of_monomials,)`
    :type coeffs: np.ndarray

    :return: result of the canonical evaluation.
    :rtype: np.ndarray
    
    """
    yy = np.dot(np.prod(np.power(pts[:, None, :], exponents[None, :, :]), axis=-1), coeffs)

    return convert_eval_output(yy)


def _verify_eval_input(pts, spatial_dimension):
    """
    verification of the input of the canonical evaluation. 
    """
    assert(isinstance(pts,np.ndarray))
    assert(pts.ndim==2)
    assert(pts.shape[-1]==spatial_dimension)


def canonical_eval(canonical_poly, pts: np.ndarray):
    """
    Navie canonical evaluation

    :param canonical_poly: Polynomial in canonical form to be evaluated.
    :type canonical_poly: CanonicalPolynomial

    :param pts: List of points, the polynomial must be evaluated on. Assumed shape: `(number_of_points,spatial_dimension)`.
    :type pts: np.ndarray

    :return: result of the canonical evaluation. 
    :rtype: np.ndarray

    """
    _verify_eval_input(pts,canonical_poly.spatial_dimension)
    return _canonical_eval(pts,canonical_poly.multi_index.exponents,canonical_poly.coeffs)


# TODO redundant
canonical_generate_internal_domain = verify_domain
canonical_generate_user_domain = verify_domain


def _canonical_partial_diff(poly: "CanonicalPolynomial", dim: int, order: int) -> "CanonicalPolynomial":
    """ Partial differentiation in Canonical basis.
    """
    spatial_dim = poly.multi_index.spatial_dimension
    deriv_order_along = np.zeros(spatial_dim, dtype=INT_DTYPE)
    deriv_order_along[dim] = order
    return _canonical_diff(poly, deriv_order_along)


def _canonical_diff(poly: "CanonicalPolynomial", order: np.ndarray) -> "CanonicalPolynomial":
    """ Partial differentiation in Canonical basis.
    """

    coeffs = poly.coeffs
    exponents = poly.multi_index.exponents

    # Guard rails in ABC ensures that the len(order) == poly.spatial_dimension
    subtracted_exponents = exponents - order

    # compute mask for non-negative multi index entries
    diff_exp_mask = np.all(exponents >= order, axis = 1)

    # multi index entries in the differentiated polynomial
    diff_exponents = subtracted_exponents[diff_exp_mask]

    # Checking if the necessary multi index entries are present
    # Zero size check is needed here as all_indices_are_contained throws error otherwise
    if diff_exponents.size != 0 and not all_indices_are_contained(diff_exponents, exponents):
        raise ValueError(f"Cannot differentiate as some of the required multi indices are not present.")

    # coefficients of the differentiated polynomial
    diff_coeffs = coeffs[diff_exp_mask] * np.prod(factorial(exponents[diff_exp_mask]) / factorial(diff_exponents),
                                                  axis=1)

    # The differentiated polynomial being expressed wrt multi indices of the given poly
    # NOTE: 'find_match_between' assumes 'exponents' is lexicographically ordered
    map_pos = find_match_between(diff_exponents, exponents)
    new_coeffs = np.zeros_like(coeffs)
    new_coeffs[map_pos] = diff_coeffs

    return CanonicalPolynomial.from_poly(poly, new_coeffs)


class CanonicalPolynomial(MultivariatePolynomialSingleABC):
    """
    Polynomial type in the canonical base.
    """

    __doc__ = """
    This is the docstring of the canonical base class.
    """

    # __doc__ += MultivariatePolynomialSingleABC.__doc_attrs__
    # Virtual Functions
    _add = staticmethod(_canonical_add)
    _sub = staticmethod(_canonical_sub)
    _mul = staticmethod(dummy)
    _div = staticmethod(dummy)
    _pow = staticmethod(dummy)
    _eval = canonical_eval

    _partial_diff = staticmethod(_canonical_partial_diff)
    _diff = staticmethod(_canonical_diff)

    generate_internal_domain = staticmethod(canonical_generate_internal_domain)
    generate_user_domain = staticmethod(canonical_generate_user_domain)



"""
The submodule `polynomials` is part of `minterpy`.

It contains concrete implementations for multivariate canonical, Newton and Lagrange polynomial base, respectively.
"""

__all__ = []

from . import canonical_polynomial  # noqa
from .canonical_polynomial import *  # noqa

__all__ += canonical_polynomial.__all__

from . import newton_polynomial  # noqa
from .newton_polynomial import *  # noqa

__all__ += newton_polynomial.__all__

from . import lagrange_polynomial  # noqa
from .lagrange_polynomial import *  # noqa

__all__ += lagrange_polynomial.__all__



"""
Module for the generating points (which provides the unisolvent nodes)
"""
from copy import deepcopy
from typing import Callable, Optional

import numpy as np

from minterpy.global_settings import ARRAY, INT_DTYPE
from minterpy.utils import gen_chebychev_2nd_order_leja_ordered

from .multi_index import MultiIndexSet
from .tree import MultiIndexTree
from .verification import check_domain_fit, check_shape, check_type_n_values

__all__ = ["Grid"]


def _gen_unisolvent_nodes(multi_index, generating_points):
    """
    .. todo::
        - document this function but ship it to utils first.
    """
    return np.take_along_axis(generating_points, multi_index.exponents, axis=0)


def get_points_from_values(spatial_dimension: int, generating_values: ARRAY):
    """
    .. todo::
        - document this function but ship it to utils first.
    """
    generating_points = np.tile(generating_values, (1, spatial_dimension))
    generating_points[:, ::2] *= -1
    return generating_points


DEFAULT_GRID_VAL_GEN_FCT = gen_chebychev_2nd_order_leja_ordered


# TODO implement comparison operations based on multi index comparison operations and the generating values used
class Grid:
    """Datatype for the nodes some polynomial bases are defined on.

    For a definition of these nodes (refered to as unisolvent nodes), see the mathematical introduction.

    .. todo::
        - insert a small introduction to the purpose of :class:`Grid` here.
        - refactor the exposed attributes (each needs at least a getter)
        - naming issues for ``generating_points`` and ``generating_values``

    Attributes
    ----------
    unisolvent_nodes
    spatial_dimension
    generating_values
    tree
    multi_index : The :class:`MultiIndexSet` instance this :class:`Grid` is based on.
    generating_points : The (multivariate) points the grid is based on.

    """

    # TODO make all attributes read only!

    _unisolvent_nodes: Optional[ARRAY] = None

    def __init__(
        self,
        multi_index: MultiIndexSet,
        generating_points: Optional[ARRAY] = None,
        generating_values: Optional[ARRAY] = None,
    ):
        if not isinstance(multi_index, MultiIndexSet):
            raise TypeError(
                f"the indices must be given as {MultiIndexSet} class instance"
            )
        # NOTE: the multi indices of a grid must be NOT be 'lexicographically complete in order to form a basis!
        # HOWEVER: building a MultiIndexTree requires complete indices
        self.multi_index: MultiIndexSet = multi_index

        if generating_points is None:
            if generating_values is None:
                generating_values = DEFAULT_GRID_VAL_GEN_FCT(multi_index.poly_degree)
            spatial_dimension = multi_index.spatial_dimension
            generating_points = get_points_from_values(
                spatial_dimension, generating_values
            )

        self.generating_values = (
            generating_values  # perform type checks and assign degree
        )

        check_type_n_values(generating_points)
        check_shape(generating_points, dimensionality=2)
        check_domain_fit(generating_points)
        self.generating_points: ARRAY = generating_points
        # TODO check if values and points fit together
        # TODO redundant information.

        self._tree: Optional[MultiIndexTree] = None

    # TODO rename: name is misleading. a generator is something different in python:
    #   cf. https://wiki.python.org/moin/Generators
    @classmethod
    def from_generator(
        cls,
        multi_index: MultiIndexSet,
        generating_function: Callable = DEFAULT_GRID_VAL_GEN_FCT,
    ):
        """
        Constructor from a factory method for the ``generating_values``.

        :param multi_index: The :class:`MultiIndexSet` this ``grid`` is based on.
        :type multi_index: MultiIndexSet

        :param generating_function: Factory method for the ``generating_values``. This functions gets a polynomial degree and returns a set of generating values of this degree.
        :type generating_function: callable

        :return: Instance of :class:`Grid` for the given input.
        :rtype: Grid


        """
        generating_values = generating_function(multi_index.poly_degree)
        return cls.from_value_set(multi_index, generating_values)

    @classmethod
    def from_value_set(cls, multi_index: MultiIndexSet, generating_values: ARRAY):
        """
        Constructor from given ``generating_values``.

        :param multi_index: The :class:`MultiIndexSet` this ``grid`` is based on.
        :type multi_index: MultiIndexSet

        :param generating_values: Generating values the :class:`Grid` instance shall be based on. The input shape needs to be one-dimensional.
        :type generating_function: np.ndarray

        :return: Instance of :class:`Grid` for the given input.
        :rtype: Grid


        """
        spatial_dimension = multi_index.spatial_dimension
        generating_points = get_points_from_values(spatial_dimension, generating_values)
        return cls(multi_index, generating_points, generating_values)

    @property
    def unisolvent_nodes(self):
        """Array of unidolvent nodes.

        For a definition of unisolvent nodes, see the mathematical introduction.

        :return: Array of the unisolvent nodes. If None were given, the output is lazily build from ``multi_index`` and ``generation_points``.
        :rtype: np.ndarray


        """
        if self._unisolvent_nodes is None:  # lazy evaluation
            self._unisolvent_nodes = _gen_unisolvent_nodes(
                self.multi_index, self.generating_points
            )
        return self._unisolvent_nodes

    @property
    def spatial_dimension(self):
        """Dimension of the domain space.

        This attribute is propagated from ``multi_index``.

        :return: The dimension of the domain space, where the polynomial will live on.
        :rtype: int

        """
        return self.multi_index.spatial_dimension

    @property
    def generating_values(self):
        """Generating values.

        A one-dimensitonal set of points, the ``generating_points`` will be build on.

        :return: Array of generating values.
        :rtype: np.ndarray
        :raise ValueError: If the given set of generating values is empty.
        :raise ValueError: If the number of points is not consistent with the polynomial degree given through ``multi_index``.

        """
        return self._generating_values

    @generating_values.setter
    def generating_values(self, values: ARRAY):
        check_type_n_values(values)
        check_domain_fit(values.reshape(-1, 1))  # 2D
        values = values.reshape(-1)  # 1D
        nr_gen_vals = len(values)
        if nr_gen_vals == 0:
            raise ValueError("at least one generating value must be given")
        self.poly_degree = nr_gen_vals - 1
        # check if multi index and generating values fit together
        if self.multi_index.poly_degree > self.poly_degree:
            raise ValueError(
                f"a grid of degree {self.poly_degree} "
                f"cannot consist of indices with degree {self.multi_index.poly_degree}"
            )
        self._unisolvent_nodes = None  # reset the unisolvent nodes
        self._generating_values: ARRAY = values

    @property
    def tree(self):
        """The used :class:`MultiIndexTree`.

        :return: The :class:`MultiIndexTree` which is connected to this :class:`Grid` instance.
        :rtype: MultiIndexTree

        .. todo::
            - is this really necessary?

        """
        if self._tree is None:  # lazy evaluation
            self._tree = MultiIndexTree(self)
        return self._tree

    def apply_func(self, func, out=None):
        """This function is not implemented yet and will raise a :class:`NotImplementedError` if called.

        Apply a given (universal) function on this :class:`Grid` instance.

        :param func: The function, which will be evaluated on the grid points.
        :type func: callable
        :raise NotImplementedError: if called, since it is not implemented yet.

        :param out: The array, where the result of the function evaluation will be stored. If given, the ``out`` array will be changed inplace, otherwise the a new one will be initialised.
        :type out: np.ndarray, optional


        .. todo::
            - implement an evaluation function for :class:`Grid` instances.
            - think about using the numpy interface for universal funcions.

        """
        # apply func to unisolvent nodes and return the func values, or store them alternatively in out
        raise NotImplementedError

    def _new_instance_if_necessary(self, multi_indices_new: MultiIndexSet) -> "Grid":
        """Constructs new grid instance only if the multi indices have changed

        :param new_indices: :class:`MultiIndexSet` instance for the ``grid``, needs to be a subset of the current ``multi_index``.
        :type new_indices: MultiIndexSet

        :return: Same :class:`Grid` instance if ``multi_index`` stays the same, otherwise new polynomial instance with the new ``multi_index``.
        :rtype: Grid
        """
        multi_indices_old = self.multi_index
        if multi_indices_new is multi_indices_old:
            return self
        # construct new:
        return self.__class__(
            multi_indices_new, self.generating_points, self.generating_values
        )

    def make_complete(self) -> "Grid":
        """completes the multi index within this :class:`Grid` instance.

        :return: completed :class:`Grid` instance
        :rtype: Grid

        Notes
        -----
        - This is required e.g. for building a multi index tree (DDS scheme)!


        """
        multi_indices_new = self.multi_index.make_complete()
        return self._new_instance_if_necessary(multi_indices_new)

    def add_points(self, exponents: ARRAY) -> "Grid":
        """Extend ``grid`` and ``multi_index``

        Adds points ``grid`` and exponents to ``multi_index`` related to a given set of additional exponents.

        :param exponents: Array of exponents added.
        :type exponents: np.ndarray

        :return: New ``grid`` with the added exponents.
        :rtype: Grid

        .. todo::
            - this is boilerplate, since similar code appears in :class:`MultivariatePolynomialSingleABC`.
        """
        exponents = np.require(exponents, dtype=INT_DTYPE)
        if np.max(exponents) > self.poly_degree:
            # TODO 'enlarge' the grid, increase the degree, ATTENTION:
            raise ValueError(
                f"trying to add point with exponent {np.max(exponents)} "
                f"but the grid is only of degree {self.poly_degree}"
            )

        multi_indices_new = self.multi_index.add_exponents(exponents)
        return self._new_instance_if_necessary(multi_indices_new)

    # copying
    def __copy__(self):
        """Creates of a shallow copy.

        This function is called, if one uses the top-level function ``copy()`` on an instance of this class.

        :return: The copy of the current instance.
        :rtype: Grid

        See Also
        --------
        copy.copy
            copy operator form the python standard library.
        """
        return self.__class__(
            self.multi_index, self.generating_points, self.generating_values
        )

    def __deepcopy__(self, mem):
        """Creates of a deepcopy.

        This function is called, if one uses the top-level function ``deepcopy()`` on an instance of this class.

        :return: The deepcopy of the current instance.
        :rtype: Grid

        See Also
        --------
        copy.deepcopy
            copy operator form the python standard library.

        """
        return self.__class__(
            deepcopy(self.multi_index),
            deepcopy(self.generating_points),
            deepcopy(self.generating_values),
        )



"""
set of utility functions to be used in polynomials submodule
"""
import itertools
import numpy as np
from minterpy.global_settings import FLOAT_DTYPE

def deriv_newt_eval(x: np.ndarray, coefficients: np.ndarray, exponents: np.ndarray,
                    generating_points: np.ndarray, derivative_order_along: np.ndarray) -> np.ndarray:
    """Evaluate the derivative of a polynomial in the Newton form.

     m = spatial dimension
     n = polynomial degree
     N = number of coefficients
     k = number of evaluation points

    Parameters
    ----------
    x: (k, m) the k points to evaluate on with dimensionality m.
    coefficients: (N) the coefficients of the Newton polynomial.
    exponents: (m, N) a multi index "alpha" for every Newton polynomial
        corresponding to the exponents of this "monomial"
    generating_points: (m, n+1) grid values for every dimension (e.g. Leja ordered Chebychev values).
    derivative_order_along: (m) specifying the order along each dimension to compute the derivative
    eg. [2,3,1] will compute respectively 2nd order, 3rd order, and 1st order along spatial dimensions
    0, 1, and 2.

    Returns
    -------
    (k) the value of derivative of the polynomial evaluated at each point.

    Notes
    -----
    - Can compute derivative polynomials without transforming to canonical basis.
    - This derivative evaluation is done by taking derivatives of the Newton monomials.
    - JIT compilation using Numba was not used here as itertools.combinations() does not work with Numba.

    """

    N, m = exponents.shape
    nr_points = x.shape[0]
    max_exponents = np.max(exponents, axis=0)

    # Result of the derivative evaluation
    results = np.empty(nr_points, dtype=FLOAT_DTYPE)
    # Array to store individial basis monomial evaluations
    monomial_vals= np.empty(N, dtype=FLOAT_DTYPE)

    num_prods = np.max(max_exponents) + 1
    # Array to store products in basis monomial along each dimension
    products = np.empty((num_prods, m), dtype=FLOAT_DTYPE)

    # Newton monomials have to be evaluated at each input point separately
    for point_nr in range(nr_points):
        x_single = x[point_nr, :]

        # Constructing the products array
        for i in range(m):
            max_exp_in_dim = max_exponents[i]
            x_i = x_single[i]
            order = derivative_order_along[i]
            if order == 0: # no partial derivative along this dimension
                prod = 1.0
                for j in range(max_exp_in_dim):  # O(n)
                    p_ij = generating_points[j, i]
                    prod *= (x_i - p_ij)
                    # NOTE: shift index by one
                    exponent = j + 1  # NOTE: otherwise the result type is float
                    products[exponent, i] = prod
            else: # take partial derivative of 'order' along this dimension

                # derivative of first 'order' newt monomials will be 0 as their degree < order
                products[:order, i] = 0.0

                # if order of derivative larger than the degree
                if order >= num_prods:
                    continue

                # derivative of newt monomial 'order' will be just factorial of order
                fact = np.math.factorial(order)
                products[order, i] = fact

                # for all bigger monomials, use chain rule of differentiation to compute derivative of products
                for q in range(order + 1, max_exp_in_dim + 1):
                    combs = itertools.combinations(range(q), q-order)
                    res = 0.0
                    for comb in combs: # combs is a generator for combinations
                        prod = np.prod(x_i - generating_points[list(comb), i])
                        res += prod

                    res *= fact
                    products[q, i] = res

        # evaluate all Newton polynomials. O(Nm)
        for j in range(N):
            # the exponents of each monomial ("alpha")
            # are the indices of the products which need to be multiplied
            newt_mon_val = 1.0  # required as multiplicative identity
            for i in range(m):
                exp = exponents[j, i]
                # NOTE: an exponent of 0 should not cause a multiplication
                if exp > 0:
                    newt_mon_val *= products[exp, i]
                else:
                    order = derivative_order_along[i]
                    if order > 0:
                        newt_mon_val = 0.0
            monomial_vals[j] = newt_mon_val

        results[point_nr] = np.dot(coefficients, monomial_vals)

    return results



"""

testing module for Transformation classes.

"""

import numpy as np
import pytest
from conftest import (
    LpDegree,
    PolyDegree,
    SpatialDimension,
    assert_call,
    assert_polynomial_almost_equal,
    build_rnd_coeffs,
)
from numpy.testing import assert_, assert_almost_equal, assert_raises

from minterpy import (
    CanonicalPolynomial,
    LagrangePolynomial,
    MultiIndexSet,
    NewtonPolynomial,
)
from minterpy.core.ABC import OperatorABC, TransformationABC
from minterpy.transformations import (
    CanonicalToLagrange,
    CanonicalToNewton,
    Identity,
    LagrangeToCanonical,
    LagrangeToNewton,
    NewtonToCanonical,
    NewtonToLagrange,
    get_transformation,
    get_transformation_class,
)
from minterpy.transformations.utils import (
    _build_lagrange_to_newton_bary,
    _build_lagrange_to_newton_naive,
    _build_newton_to_lagrange_bary,
    _build_newton_to_lagrange_naive,
    build_l2n_matrix_dds,
)

transform_classes = [
    LagrangeToNewton,
    NewtonToLagrange,
    LagrangeToCanonical,
    CanonicalToLagrange,
    NewtonToCanonical,
    CanonicalToNewton,
]


@pytest.fixture(params=transform_classes)
def Transform(request):
    return request.param


def test_init_transform(Transform):
    """testing the initialization of transformation classes"""

    assert_(issubclass(Transform, TransformationABC))

    # Test initialization
    mi = MultiIndexSet.from_degree(2, 2, 1.0)
    coeffs = np.arange(len(mi), dtype=float)
    poly = Transform.origin_type(mi, coeffs)
    assert_call(Transform, poly)

    # test transformation call
    transform = Transform(poly)
    assert_call(transform)

    # test the type of transformed poly
    res_poly = transform()
    assert_(isinstance(res_poly, Transform.target_type))

    # test existence of transformation_operator
    operator = transform.transformation_operator
    assert_(isinstance(operator, OperatorABC))


poly_classes = [CanonicalPolynomial, NewtonPolynomial, LagrangePolynomial]


@pytest.fixture(params=poly_classes)
def Polynom(request):
    return request.param


P1 = Polynom
P2 = Polynom


def test_get_transformation(P1, P2):
    """ test the get_transformation function in transformation_meta"""

    mi = MultiIndexSet.from_degree(2, 1, 1.0)
    coeffs = np.arange(len(mi), dtype=float)
    poly = P1(mi, coeffs)

    transform = get_transformation(poly, P2)

    if P1 == P2:
        assert_(isinstance(transform, Identity))
    else:
        assert_(isinstance(transform, TransformationABC))


def test_fail_get_transformation_class():
    """ tests if get_transformation_class throws an error if it cannot find a transforamtion"""

    assert_raises(
        NotImplementedError, get_transformation_class, None, LagrangePolynomial
    )


def test_l2n_transform(SpatialDimension, PolyDegree, LpDegree):
    """Tests the Lagrange to Newton transformations."""

    # Arrange
    mi = MultiIndexSet.from_degree(SpatialDimension, PolyDegree, LpDegree)
    coeffs = build_rnd_coeffs(mi)
    lag_poly = LagrangePolynomial(mi, coeffs)

    transformation_l2n = LagrangeToNewton(lag_poly)

    # Test the naive form
    transform_naive = _build_lagrange_to_newton_naive(transformation_l2n)
    newt_coeffs_naive = transform_naive @ lag_poly.coeffs
    newt_poly_naive = NewtonPolynomial(mi, newt_coeffs_naive)

    # Test the barycentric form
    transform_bary = _build_lagrange_to_newton_bary(transformation_l2n)
    newt_coeffs_bary = transform_bary @ lag_poly.coeffs
    newt_poly_bary = NewtonPolynomial(mi, newt_coeffs_bary)

    # Compare the results of the naive and barycentric transformation
    assert_polynomial_almost_equal(newt_poly_naive, newt_poly_bary)

    # Test the transformation via DDS of the Lagrange polynomial on the grid
    l2n_matrix = build_l2n_matrix_dds(lag_poly.grid)

    # Compare the results of the naive and DDS transformation
    assert_almost_equal(l2n_matrix, transform_naive.array_repr_full)

    # Test if the Newton interpolation polynomials evaluated on the unisolvent
    # nodes are indeed the Lagrange coefficients
    res_eval = newt_poly_naive(lag_poly.unisolvent_nodes)
    assert_almost_equal(res_eval, lag_poly.coeffs)


def test_n2l_transform(SpatialDimension, PolyDegree, LpDegree):
    """ testing the naive and bary centric l2n transformations """
    mi = MultiIndexSet.from_degree(SpatialDimension, PolyDegree, LpDegree)
    coeffs = build_rnd_coeffs(mi)
    newt_poly = NewtonPolynomial(mi, coeffs)

    transformation_n2l = NewtonToLagrange(newt_poly)

    # test naive
    transform_naive = _build_newton_to_lagrange_naive(transformation_n2l)
    lag_coeffs_naive = transform_naive @ newt_poly.coeffs
    lag_poly_naive = LagrangePolynomial(mi, lag_coeffs_naive)

    # test bary dict
    transform_bary = _build_newton_to_lagrange_bary(transformation_n2l)
    lag_coeffs_bary = transform_bary @ newt_poly.coeffs
    lag_poly_bary = LagrangePolynomial(mi, lag_coeffs_bary)

    # compare the result of naive and bary transformation
    assert_polynomial_almost_equal(lag_poly_naive, lag_poly_bary)


def test_transformation_identity():

    mi = MultiIndexSet.from_degree(2, 1, 1.0)
    coeffs = build_rnd_coeffs(mi)
    newt_poly = NewtonPolynomial(mi, coeffs)

    assert_call(Identity, newt_poly)
    transform = Identity(newt_poly)

    transform_mat = transform.transformation_operator.array_repr_full

    # check whether transformation matrix is identity
    assert_almost_equal(transform_mat, np.eye(len(mi)))

    assert_call(transform)
    res_poly = transform()

    # check the resulting polynomial is same after transformation
    assert_polynomial_almost_equal(res_poly, newt_poly)


def test_transform_back_n_forth(P1, P2, SpatialDimension, PolyDegree, LpDegree):
    mi = MultiIndexSet.from_degree(SpatialDimension, PolyDegree, LpDegree)
    coeffs = build_rnd_coeffs(mi)
    origin_poly = P1(mi, coeffs)

    transform_forward = get_transformation(origin_poly, P2)
    interim_poly = transform_forward()

    transform_back = get_transformation(interim_poly, P1)
    final_poly = transform_back()

    assert_polynomial_almost_equal(origin_poly, final_poly)



"""
The submodule `ABC` is part of `minterpy`.

It contains all AbstractBaseClasses used in `minterpy`.
"""

__all__ = []


from . import multivariate_polynomial_abstract  # noqa
from .multivariate_polynomial_abstract import *  # noqa

__all__ += multivariate_polynomial_abstract.__all__


from . import transformation_abstract  # noqa
from .transformation_abstract import *  # noqa

__all__ += transformation_abstract.__all__


from . import operator_abstract  # noqa
from .operator_abstract import *  # noqa

__all__ += operator_abstract.__all__



"""
testing module for canonical_polynomial.py

The subclassing is not tested here, see tesing module `test_polynomial.py`
"""
import operator as op

import numpy as np
import pytest
from conftest import (
    SEED,
    LpDegree,
    MultiIndices,
    NrPoints,
    NrSimilarPolynomials,
    PolyDegree,
    SpatialDimension,
    assert_polynomial_almost_equal,
    build_rnd_coeffs,
    build_rnd_points,
)
from numpy.testing import assert_, assert_almost_equal

from minterpy import CanonicalPolynomial, MultiIndexSet

# tests with a single polynomial


def test_neg(MultiIndices, NrSimilarPolynomials):
    coeffs = build_rnd_coeffs(MultiIndices, NrSimilarPolynomials)
    poly = CanonicalPolynomial(MultiIndices, coeffs)
    res = -poly
    groundtruth_coeffs = (-1) * coeffs
    groundtruth = poly.__class__(MultiIndices, groundtruth_coeffs)
    assert_polynomial_almost_equal(res, groundtruth)


def test_eval(MultiIndices, NrPoints):
    coeffs = build_rnd_coeffs(MultiIndices)
    poly = CanonicalPolynomial(MultiIndices, coeffs)
    pts = build_rnd_points(NrPoints, MultiIndices.spatial_dimension)
    res = poly(pts)

    # navie impementation of canonical eval
    # related to issue #32
    groundtruth = np.zeros(NrPoints)
    for k,pt in enumerate(pts):
        single_groundtruth = 0.0
        for i,exponents in enumerate(MultiIndices.exponents):
            term = 1.0
            for j, expo in enumerate(exponents):
                term *= pt[j]**expo
            single_groundtruth+=coeffs[i]*term
        groundtruth[k] = single_groundtruth

    assert_almost_equal(res, groundtruth)


# tests with two polynomials
# todo:: find out if there are some more sophisticated tests for that
exps1 = np.array([[0, 0], [1, 0], [0, 1], [1, 1]])
args1 = np.lexsort(exps1.T, axis=-1)
mi1 = MultiIndexSet(exps1[args1])
coeffs1 = np.array([1, 2, 3, 4])

exps2 = np.array([[0, 0, 0], [2, 0, 0], [1, 1, 0], [0, 2, 0], [0, 0, 2]])
args2 = np.lexsort(exps2.T, axis=-1)
mi2 = MultiIndexSet(exps2[args2])
coeffs2 = np.array([1, 2, 3, 4, 5])

polys = [
    CanonicalPolynomial(mi1, coeffs1[args1]),
    CanonicalPolynomial(mi2, coeffs2[args2]),
]


@pytest.fixture(params=polys)
def Poly(request):
    return request.param


def test_sub_same_poly(Poly):
    res = Poly - Poly
    groundtruth_coeffs = np.zeros(Poly.coeffs.shape)
    groundtruth = Poly.__class__(Poly.multi_index, groundtruth_coeffs)
    assert_polynomial_almost_equal(res, groundtruth)


P1 = Poly
P2 = Poly


def test_add_different_poly(P1, P2):
    res = P1 + P2
    if P1 is P2:
        groundtruth_coeffs = P1.coeffs * 2
        groundtruth_multi_index_exponents = P1.multi_index.exponents
    else:
        groundtruth_coeffs = np.array([2, 2, 2, 3, 7, 4, 5])
        groundtruth_multi_index_exponents = np.array(
            [
                [0, 0, 0],
                [1, 0, 0],
                [2, 0, 0],
                [0, 1, 0],
                [1, 1, 0],
                [0, 2, 0],
                [0, 0, 2],
            ]
        )
    groundtruth_multi_index = MultiIndexSet(groundtruth_multi_index_exponents)
    groundtruth = P1.__class__(groundtruth_multi_index, groundtruth_coeffs)
    assert_polynomial_almost_equal(res, groundtruth)


def test_sub_different_poly():
    """
    .. todo::
        - make this a bit better
    """
    res = polys[0] - polys[1]
    groundtruth_coeffs = np.array([0, 2, -2, 3, 1, -4, -5])
    groundtruth_multi_index_exponents = np.array(
        [
            [0, 0, 0],
            [1, 0, 0],
            [2, 0, 0],
            [0, 1, 0],
            [1, 1, 0],
            [0, 2, 0],
            [0, 0, 2],
        ]
    )
    groundtruth_multi_index = MultiIndexSet(groundtruth_multi_index_exponents)
    groundtruth = polys[0].__class__(groundtruth_multi_index, groundtruth_coeffs)
    assert_polynomial_almost_equal(res, groundtruth)


def test_partial_diff():

    # ATTENTION: the exponent vectors of all derivatives have to be included already!
    exponents = np.array([[0, 0, 0],
                          [0, 1, 0],
                          [0, 0, 1],
                          [0, 1, 1],
                          [0, 0, 2]])
    coeffs = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
    assert exponents.shape == (5, 3)
    assert coeffs.shape == (5,)

    mi = MultiIndexSet(exponents)
    can_poly = CanonicalPolynomial(mi, coeffs)

    groundtruth_coeffs_dx = np.array([0.0, 0.0, 0.0, 0.0, 0.0])
    groundtruth_coeffs_dy = np.array([2.0, 0.0, 4.0, 0.0, 0.0])
    groundtruth_coeffs_dz = np.array([3.0, 4.0, 10.0, 0.0, 0.0])

    can_poly_dx = can_poly.partial_diff(0)
    coeffs_dx = can_poly_dx.coeffs
    assert np.allclose(coeffs_dx, groundtruth_coeffs_dx)

    can_poly_dy = can_poly.partial_diff(1)
    coeffs_dy = can_poly_dy.coeffs
    assert np.allclose(coeffs_dy, groundtruth_coeffs_dy)

    can_poly_dz = can_poly.partial_diff(2)
    coeffs_dz = can_poly_dz.coeffs
    assert np.allclose(coeffs_dz, groundtruth_coeffs_dz)



def test_diff():

    # ATTENTION: the exponent vectors of all derivatives have to be included already!
    exponents = np.array([[0, 0, 0],
                          [0, 1, 0],
                          [0, 0, 1],
                          [0, 1, 1],
                          [0, 0, 2]])
    coeffs = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
    assert exponents.shape == (5, 3)
    assert coeffs.shape == (5,)

    mi = MultiIndexSet(exponents)
    can_poly = CanonicalPolynomial(mi, coeffs)

    # Testing zeroth order derivatives
    can_poly_zero_deriv = can_poly.diff([0,0,0])
    coeffs_zero_deriv = can_poly_zero_deriv.coeffs
    assert np.allclose(coeffs_zero_deriv, coeffs)

    groundtruth_coeffs_dyz = np.array([4.0, 0.0, 0.0, 0.0, 0.0])
    groundtruth_coeffs_dz2 = np.array([10.0, 0.0, 0.0, 0.0, 0.0])
    groundtruth_coeffs_dyz2 = np.array([0.0, 0.0, 0.0, 0.0, 0.0])

    can_poly_dyz = can_poly.diff([0,1,1])
    coeffs_dyz = can_poly_dyz.coeffs
    assert np.allclose(coeffs_dyz, groundtruth_coeffs_dyz)

    can_poly_dz2 = can_poly.diff([0,0,2])
    coeffs_dz2 = can_poly_dz2.coeffs
    assert np.allclose(coeffs_dz2, groundtruth_coeffs_dz2)

    can_poly_dyz2 = can_poly.diff([0,1,2])
    coeffs_dyz2 = can_poly_dyz2.coeffs
    assert np.allclose(coeffs_dyz2, groundtruth_coeffs_dyz2)



"""
Testing module for interpolation.py

Here the functionality of the respective attribute is not tested.
"""

import numpy as np
import pytest
from conftest import (
    LpDegree,
    MultiIndices,
    NrPoints,
    PolyDegree,
    SpatialDimension,
    assert_call,
    assert_grid_equal,
    assert_interpolant_almost_equal,
    assert_multi_index_equal,
    assert_polynomial_almost_equal,
    build_random_newton_polynom,
    build_rnd_points,
)
from numpy.testing import assert_, assert_almost_equal

import minterpy as mp
from minterpy import Interpolant, Interpolator, interpolate

# test construction


def test_init_interpolator(SpatialDimension, PolyDegree, LpDegree):
    assert_call(Interpolator, SpatialDimension, PolyDegree, LpDegree)
    interpolator = Interpolator(SpatialDimension, PolyDegree, LpDegree)
    groundtruth_multi_index = mp.MultiIndexSet.from_degree(
        SpatialDimension, PolyDegree, LpDegree
    )
    groundtruth_grid = mp.Grid(groundtruth_multi_index)
    assert_multi_index_equal(interpolator.multi_index, groundtruth_multi_index)
    assert_grid_equal(interpolator.grid, groundtruth_grid)


def test_init_interpolant(SpatialDimension, PolyDegree, LpDegree):
    assert_call(
        Interpolant,
        lambda x: x[:, 0],
        Interpolator(SpatialDimension, PolyDegree, LpDegree),
    )
    assert_call(
        Interpolant.from_degree,
        lambda x: x[:, 0],
        SpatialDimension,
        PolyDegree,
        LpDegree,
    )
    interpolant_default = Interpolant(
        lambda x: x[:, 0], Interpolator(SpatialDimension, PolyDegree, LpDegree)
    )
    interpolant_from_degree = Interpolant.from_degree(
        lambda x: x[:, 0], SpatialDimension, PolyDegree, LpDegree
    )
    assert_interpolant_almost_equal(interpolant_default, interpolant_from_degree)


def test_call_interpolate(SpatialDimension, PolyDegree, LpDegree):
    assert_call(interpolate, lambda x: x[:, 0], SpatialDimension, PolyDegree, LpDegree)
    interpolant = interpolate(lambda x: x[:, 0], SpatialDimension, PolyDegree, LpDegree)
    assert_(isinstance(interpolant, Interpolant))


# test if the interpolator can interpolate
def test_interpolator(SpatialDimension, PolyDegree, LpDegree):
    groundtruth_poly = build_random_newton_polynom(
        SpatialDimension, PolyDegree, LpDegree
    )
    interpolator = Interpolator(SpatialDimension, PolyDegree, LpDegree)
    res_from_newton_poly = interpolator(groundtruth_poly)
    res_from_canonical_poly = interpolator(mp.NewtonToCanonical(groundtruth_poly)())
    assert_polynomial_almost_equal(res_from_newton_poly, groundtruth_poly)
    assert_polynomial_almost_equal(res_from_canonical_poly, groundtruth_poly)


# test if the interpolant interpolates
def test_interpolant(NrPoints, SpatialDimension, PolyDegree, LpDegree):
    rnd_points = build_rnd_points(NrPoints, SpatialDimension)
    groundtruth_poly = build_random_newton_polynom(
        SpatialDimension, PolyDegree, LpDegree
    )
    groundtruth = groundtruth_poly(rnd_points)
    interpolant = Interpolant.from_degree(
        groundtruth_poly, SpatialDimension, PolyDegree, LpDegree
    )
    res = interpolant(rnd_points)
    assert_almost_equal(res, groundtruth)


# test if the interpolate does what it promisses
def test_interpolate(NrPoints, SpatialDimension, PolyDegree, LpDegree):
    rnd_points = build_rnd_points(NrPoints, SpatialDimension)
    groundtruth_poly = build_random_newton_polynom(
        SpatialDimension, PolyDegree, LpDegree
    )
    groundtruth = groundtruth_poly(rnd_points)
    interpolant = interpolate(groundtruth_poly, SpatialDimension, PolyDegree, LpDegree)
    res = interpolant(rnd_points)
    assert_almost_equal(res, groundtruth)



"""
Test suite for multi_index_utils.py
"""
import numpy as np
import pytest
from conftest import (
    LpDegree,
    PolyDegree,
    SpatialDimension,
    assert_call,
    build_rnd_exponents,
)
from numpy.testing import assert_, assert_equal, assert_raises

from minterpy.core.utils import (
    _gen_multi_index_exponents,
    find_match_between,
    get_exponent_matrix,
    insert_lexicographically,
    is_lexicographically_complete,
    make_complete,
    make_derivable,
)
from minterpy.jit_compiled_utils import (
    all_indices_are_contained,
    have_lexicographical_ordering,
    index_is_contained,
    lex_smaller_or_equal,
)

MIN_POLY_DEG = 1
MAX_POLY_DEG = 5
SEED = 12345678


def get_random_dim(spatial_dimension, seed=None):
    if seed is None:
        seed = SEED
    if spatial_dimension == 1:
        rnd_dim = 0
    else:
        rnd_dim = np.random.randint(0, spatial_dimension - 1)
    return rnd_dim


def get_lex_bigger(index, bigger_by_1=False, seed=None):
    rnd_dim = get_random_dim(len(index))
    out = index.copy()
    out[rnd_dim] += 1
    # setting all previous entries to 0 does not change the lexicographical ordering
    if not bigger_by_1:
        out[:rnd_dim] = 0
    return out


# fixtures for number of monomials
number_of_monomials = [1, 2]


@pytest.fixture(params=number_of_monomials)
def NumOfMonomials(request):
    return request.param


# test utilities

def test_call_get_exponent_matrix(SpatialDimension, PolyDegree, LpDegree):
    
    # The function can be called
    assert_call(get_exponent_matrix, SpatialDimension, PolyDegree, LpDegree)

    # The function can be called with finite integral float as poly. degree
    # NOTE: This test is related to the fix for Issue #65
    poly_degree = float(PolyDegree)
    assert_call(get_exponent_matrix, SpatialDimension, poly_degree, LpDegree)
    
    # The function can't be called non-finite integral float
    # NOTE: This test is related to the fix for Issue #65
    poly_degree = PolyDegree + np.random.rand(1)[0]
    assert_raises(
        ValueError,
        get_exponent_matrix,
        SpatialDimension,
        poly_degree,
        LpDegree
        )


def test_index_is_contained(SpatialDimension, PolyDegree, LpDegree):
    exponents = get_exponent_matrix(SpatialDimension, PolyDegree, LpDegree)
    number_of_monomials, dim = exponents.shape
    assert_(dim == SpatialDimension)

    # TODO: is it necessary to do this for every element?
    for exponent_vector in exponents:
        assert_(index_is_contained(exponents, exponent_vector))

    largest_exponent_vector = exponents[-1, :]  # last / biggest exponent vector
    bigger_exponent_vector = get_lex_bigger(largest_exponent_vector)
    assert_(not index_is_contained(exponents, bigger_exponent_vector))

    deleted_exponent_vector = exponents[0]
    exponents2 = np.delete(exponents, 0, axis=0)
    assert_(not index_is_contained(exponents2, deleted_exponent_vector))


def test_lex_smaller_or_equal(SpatialDimension, PolyDegree, LpDegree):
    exponents = get_exponent_matrix(SpatialDimension, PolyDegree, LpDegree)
    number_of_monomials, dim = exponents.shape

    # TODO: is it necessary to do this for every element?
    for exp_vect in exponents:
        assert_(lex_smaller_or_equal(exp_vect, exp_vect))
        bigger_exponent_vector = get_lex_bigger(exp_vect)
        assert_(lex_smaller_or_equal(exp_vect, bigger_exponent_vector))
        assert_(not lex_smaller_or_equal(bigger_exponent_vector, exp_vect))


def test_is_lexicographically_complete(SpatialDimension, PolyDegree, LpDegree):
    exponents = get_exponent_matrix(SpatialDimension, PolyDegree, LpDegree)
    assert_(is_lexicographically_complete(exponents))

    # NOTE: Only applies for PolyDegree > 0 (== 0 has nothing to remove)
    if PolyDegree > 0:
        # deleting the smallest vector destroys completeness
        assert_(not is_lexicographically_complete(np.delete(exponents, 0, axis=0)))

        # deleting the first row of exponents of PolyDegree > 1
        if PolyDegree > 1:
            assert_(
                not is_lexicographically_complete(
                    np.delete(exponents, 1, axis=0)
                )
            )

        # deleting the biggest vector maintains completeness
        assert_(is_lexicographically_complete(np.delete(exponents, -1, axis=0)))

    # TODO: shall not use functions which are not tested!
    # bigger_exponent_vector = expomemts[-1,:].copy()  # independent copy!
    # bigger_exponent_vector[0] += 2  # introduces "hole"
    # multi_index2 = multi_index.add_exponents(bigger_exponent_vector)
    # exponents2 = multi_index2.exponents
    # assert not is_lexicographically_complete(exponents2)
    # if multi_index2.is_complete:
    # assert not multi_index2.is_complete


def test_insert_indices(SpatialDimension, PolyDegree, LpDegree):
    exponents = get_exponent_matrix(SpatialDimension, PolyDegree, LpDegree)
    number_of_monomials, dim = exponents.shape

    # TODO: is it necessary to do this for every element?
    for idx, exp_vect in enumerate(exponents):

        exponents2 = insert_lexicographically(exponents, exp_vect)
        assert_equal(exponents2, exponents)

        incomplete_exponents = np.delete(exponents, idx, axis=0)
        restored_exponents = insert_lexicographically(exponents, exp_vect)
        assert_equal(restored_exponents, exponents)

    # bigger indices should be inserted at the end
    largest_exponent_vector = exponents[-1, :]  # last / biggest exponent vector
    bigger_exponent_vector = get_lex_bigger(largest_exponent_vector)
    enlarged_exponents = insert_lexicographically(exponents, bigger_exponent_vector)
    assert_(enlarged_exponents.shape[0] == (number_of_monomials + 1))
    assert_(enlarged_exponents.shape[1] == dim)
    assert_equal(enlarged_exponents[:-1], exponents)


def test_make_complete(SpatialDimension, PolyDegree, LpDegree):
    exponents = get_exponent_matrix(SpatialDimension, PolyDegree, LpDegree)
    assert_(is_lexicographically_complete(exponents))

    # Remove the first row and then make the exponents complete
    # NOTE: Only applies for PolyDegree > 0
    if PolyDegree > 0:
        exp_vect = exponents[0]
        incomplete_exponents = np.delete(exponents, 0, axis=0)
        # completion should be identical to "make derivable"
        # after just deleting a single exponent vector
        completed_exponents1 = make_derivable(incomplete_exponents)
        completed_exponents2 = make_complete(incomplete_exponents, LpDegree)

        assert is_lexicographically_complete(completed_exponents1)
        assert is_lexicographically_complete(completed_exponents2)

        assert_equal(exp_vect, completed_exponents1[0, :])
        assert_equal(exp_vect, completed_exponents2[0, :])

    # Remove the second row and the make the exponents complete
    # NOTE: Only applies for PolyDegree > 1
    #       (otherwise, the exponents are still complete)
    if PolyDegree > 1:
        exp_vect = exponents[1]
        incomplete_exponents = np.delete(exponents, 1, axis=0)
        # completion should be identical to "make derivable"
        # after just deleting a single exponent vector
        completed_exponents1 = make_derivable(incomplete_exponents)
        completed_exponents2 = make_complete(incomplete_exponents, LpDegree)

        assert is_lexicographically_complete(completed_exponents1)
        assert is_lexicographically_complete(completed_exponents2)

        assert_equal(exp_vect, completed_exponents1[1, :])
        assert_equal(exp_vect, completed_exponents2[1, :])

    # Insertion of a term that belongs to a higher-degree set
    # NOTE: This test is related to the fix for Issue #65
    incomplete_exponents = np.insert(
        exponents, len(exponents), exponents[-1]+2, axis=0
    )
    # Make sure that the incomplete set is indeed incomplete
    assert_(not is_lexicographically_complete(incomplete_exponents))
    # Make the set complete
    completed_exponents = make_complete(incomplete_exponents, LpDegree)
    # Completed set must be lexicographically ordered
    assert_(have_lexicographical_ordering(completed_exponents))
    # Completed set must be complete
    assert_(is_lexicographically_complete(completed_exponents))


def test_all_indices_are_contained(SpatialDimension, PolyDegree, LpDegree):
    exponents = get_exponent_matrix(SpatialDimension, PolyDegree, LpDegree)

    # NOTE: Only applies for PolyDegree > 0 (== 0 has nothing to remove)
    # TODO: is it necessary to do this for every element?
    if PolyDegree > 0:
        for idx in range(exponents.shape[0]):
            incomplete_exponents = np.delete(exponents, idx, axis=0)
            assert_(
                not all_indices_are_contained(
                    exponents, incomplete_exponents
                )
            )
            assert_(all_indices_are_contained(incomplete_exponents, exponents))

    # when adding indices, not all indices are contained <-- ???
    largest_exponent_vector = exponents[-1, :]  # last / biggest exponent vector
    bigger_exponent_vector = get_lex_bigger(largest_exponent_vector)
    enlarged_exponents = insert_lexicographically(
        exponents, bigger_exponent_vector
    )
    assert_(not all_indices_are_contained(enlarged_exponents, exponents))
    # but the other way round should hold:
    assert_(all_indices_are_contained(exponents, enlarged_exponents))



"""
Concrete implementations of the Transformation classes for the NewtonPolynomial.

Transformations from Newton basis to Canonical and Lagrange basis are provided.
"""

from minterpy.core.ABC import TransformationABC
from minterpy.polynomials import CanonicalPolynomial, LagrangePolynomial
from minterpy.polynomials.newton_polynomial import NewtonPolynomial

from .utils import (
    _build_newton_to_canonical_operator,
    _build_newton_to_lagrange_operator,
)

__all__ = ["NewtonToCanonical", "NewtonToLagrange"]


class NewtonToCanonical(TransformationABC):
    """Transformation from NewtonPolynomial to CanonicalPolynomial"""

    origin_type = NewtonPolynomial
    target_type = CanonicalPolynomial
    _get_transformation_operator = _build_newton_to_canonical_operator


class NewtonToLagrange(TransformationABC):
    """Transformation from NewtonPolynomial to LagrangePolynomial"""

    origin_type = NewtonPolynomial
    target_type = LagrangePolynomial
    _get_transformation_operator = _build_newton_to_lagrange_operator



"""
Testing module for lagrange_polynomial.py

"""

import numpy as np
from conftest import (
    MultiIndices,
    NrPoints,
    NrSimilarPolynomials,
    assert_polynomial_almost_equal,
    build_rnd_coeffs,
    build_rnd_points,
)

from minterpy import CanonicalPolynomial, LagrangePolynomial, MultiIndexSet
from minterpy.transformations import CanonicalToLagrange, LagrangeToCanonical


def test_neg(MultiIndices, NrSimilarPolynomials):
    coeffs = build_rnd_coeffs(MultiIndices, NrSimilarPolynomials)
    poly = LagrangePolynomial(MultiIndices, coeffs)
    res = -poly
    groundtruth_coeffs = (-1) * coeffs
    groundtruth = poly.__class__(MultiIndices, groundtruth_coeffs)
    assert_polynomial_almost_equal(res, groundtruth)


Mi1 = MultiIndices
Mi2 = MultiIndices


def test_add_poly(Mi1, Mi2):
    coeffs1 = build_rnd_coeffs(Mi1)
    coeffs2 = build_rnd_coeffs(Mi2)

    lag_poly_1 = LagrangePolynomial(Mi1, coeffs1)
    lag_poly_2 = LagrangePolynomial(Mi2, coeffs2)

    res = lag_poly_1 + lag_poly_2

    transform_l2c_poly_1 = LagrangeToCanonical(lag_poly_1)
    transform_l2c_poly_2 = LagrangeToCanonical(lag_poly_2)
    transform_l2c_res = LagrangeToCanonical(res)

    can_poly_1 = transform_l2c_poly_1()
    can_poly_2 = transform_l2c_poly_2()
    can_poly_res = transform_l2c_res()

    groundtruth_res_poly = can_poly_1 + can_poly_2

    assert_polynomial_almost_equal(groundtruth_res_poly, can_poly_res)


def test_sub_poly(Mi1, Mi2):
    coeffs1 = build_rnd_coeffs(Mi1)
    coeffs2 = build_rnd_coeffs(Mi2)

    lag_poly_1 = LagrangePolynomial(Mi1, coeffs1)
    lag_poly_2 = LagrangePolynomial(Mi2, coeffs2)

    res = lag_poly_1 - lag_poly_2

    transform_l2c_poly_1 = LagrangeToCanonical(lag_poly_1)
    transform_l2c_poly_2 = LagrangeToCanonical(lag_poly_2)
    transform_l2c_res = LagrangeToCanonical(res)

    can_poly_1 = transform_l2c_poly_1()
    can_poly_2 = transform_l2c_poly_2()
    can_poly_res = transform_l2c_res()

    groundtruth_res_poly = can_poly_1 - can_poly_2
    assert_polynomial_almost_equal(can_poly_res, groundtruth_res_poly)


def test_mul():
    mi = MultiIndexSet.from_degree(2, 1, 2.0)
    canonical_coeffs1 = np.array([0, 1, 1])
    canonical_coeffs2 = np.array([0, 1, -1])

    can_poly_1 = CanonicalPolynomial(mi, canonical_coeffs1)
    can_poly_2 = CanonicalPolynomial(mi, canonical_coeffs2)

    c2l_transformation = CanonicalToLagrange(can_poly_1)

    lagrange_poly1 = c2l_transformation(can_poly_1)
    lagrange_poly2 = c2l_transformation(can_poly_2)

    res_lagrange_poly = lagrange_poly1 * lagrange_poly2

    l2c_transformation = LagrangeToCanonical(res_lagrange_poly)
    res_canonical_poly = l2c_transformation()

    groundtruth_product_coeffs = np.array([0, 0, 1, 0, 0, -1])
    groundtruth_mi = MultiIndexSet.from_degree(2, 2, 4.0)
    groundtruth = CanonicalPolynomial(groundtruth_mi, groundtruth_product_coeffs)

    assert_polynomial_almost_equal(groundtruth, res_canonical_poly)


def test_add_different_poly():
    mi1 = MultiIndexSet.from_degree(3, 2, 2.0)
    coeffs1 = np.array([1, 2, 0, 4, 3, 0, 5, 0, 0, 0, 0])

    mi2 = MultiIndexSet.from_degree(2, 2, 2.0)
    coeffs2 = np.array([1, 2, 0, 3, 4, 0])

    can_poly_1 = CanonicalPolynomial(mi1, coeffs1)
    can_poly_2 = CanonicalPolynomial(mi2, coeffs2)

    transform_c2l_1 = CanonicalToLagrange(can_poly_1)
    transform_c2l_2 = CanonicalToLagrange(can_poly_2)

    lag_poly_1 = transform_c2l_1()
    lag_poly_2 = transform_c2l_2()

    res = lag_poly_1 + lag_poly_2

    groundtruth_coeffs = np.array([2, 4, 0, 7, 7, 0, 5, 0, 0, 0, 0])

    groundtruth_multi_index = mi1

    groundtruth_canonical = CanonicalPolynomial(
        groundtruth_multi_index, groundtruth_coeffs
    )

    transform_c2l_res = CanonicalToLagrange(groundtruth_canonical)
    groundtruth = transform_c2l_res()

    assert_polynomial_almost_equal(res, groundtruth)



"""
Concrete implementations of the Transformation classes for the CanonicalPolynomial.

Transformations from Canonical basis to Newton and Lagrange basis are provided.
"""

from minterpy.core.ABC import TransformationABC
from minterpy.polynomials import CanonicalPolynomial
from minterpy.polynomials.lagrange_polynomial import LagrangePolynomial
from minterpy.polynomials.newton_polynomial import NewtonPolynomial

from .utils import (
    _build_canonical_to_lagrange_operator,
    _build_canonical_to_newton_operator,
)

__all__ = ["CanonicalToNewton", "CanonicalToLagrange"]


class CanonicalToNewton(TransformationABC):
    """Transformation from CanonicalPolynomial to NewtonPolynomial"""

    origin_type = CanonicalPolynomial
    target_type = NewtonPolynomial
    _get_transformation_operator = _build_canonical_to_newton_operator


class CanonicalToLagrange(TransformationABC):
    """Transformation from CanonicalPolynomial to LagrangePolynomial"""

    origin_type = CanonicalPolynomial
    target_type = LagrangePolynomial
    _get_transformation_operator = _build_canonical_to_lagrange_operator



#!/usr/bin/env python
# Copyright (c) 2021, Uwe Hernandez Acosta
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://gitlab.hzdr.de/interpol/minterpy for details.

from setuptools import setup

setup()



"""
LagrangePolynomial class
"""
from typing import Any, Optional

import numpy as np

import minterpy
from minterpy.global_settings import ARRAY

from ..core import Grid, MultiIndexSet
from ..core.ABC import MultivariatePolynomialSingleABC
from ..core.verification import verify_domain
from .canonical_polynomial import _match_dims, _matching_internal_domain

__all__ = ["LagrangePolynomial"]


def dummy(x: Optional[Any] = None) -> None:
    """Placeholder function.

    .. warning::
      This feature is not implemented yet!
    """
    raise NotImplementedError("This feature is not implemented yet.")


# TODO : poly2 can be of a different basis?
def _lagrange_add(
    poly1: MultivariatePolynomialSingleABC, poly2: MultivariatePolynomialSingleABC
) -> MultivariatePolynomialSingleABC:
    """Addition of two polynomials in Lagrange basis.


    :param poly1: First polynomial to be added.
    :type poly1: LagrangePolynomial
    :param poly2: Second polynomial to be added.
    :type poly2: LagrangePolynomial

    :return: The result of ``poly1 + poly2``
    :rtype: LagrangePolynomial

    """
    p1, p2 = _match_dims(poly1, poly2)
    if _matching_internal_domain(p1, p2):
        l2n_p1 = minterpy.transformations.LagrangeToNewton(p1)
        newt_p1 = l2n_p1()
        l2n_p2 = minterpy.transformations.LagrangeToNewton(p2)
        newt_p2 = l2n_p2()

        max_poly_degree = np.max(
            np.array([p1.multi_index.poly_degree, p2.multi_index.poly_degree])
        )
        max_lp_degree = np.max(
            np.array([p1.multi_index.lp_degree, p2.multi_index.lp_degree])
        )

        dim = p1.spatial_dimension  # must be the same for p2

        res_mi = MultiIndexSet.from_degree(dim, int(max_poly_degree), max_lp_degree)
        res_grid = Grid(res_mi)

        un = res_grid.unisolvent_nodes

        eval_p1 = newt_p1(un)
        eval_p2 = newt_p2(un)

        res_coeffs = eval_p1 + eval_p2

        return LagrangePolynomial(
            res_mi,
            res_coeffs,
            grid=res_grid,
            internal_domain=p1.internal_domain,
            user_domain=p1.user_domain,
        )
    else:
        raise NotImplementedError(
            "Addition is not implemented for lagrange polynomials with different domains"
        )


def _lagrange_sub(
    poly1: MultivariatePolynomialSingleABC, poly2: MultivariatePolynomialSingleABC
) -> MultivariatePolynomialSingleABC:
    """Subtraction of two polynomials in Lagrange basis.


    :param poly1: First polynomial from which will be substracted.
    :type poly1: LagrangePolynomial
    :param poly2: Second polynomial which is substracted.
    :type poly2: LagrangePolynomial

    :return: The result of ``poly1 - poly2``
    :rtype: LagrangePolynomial
    """
    p1, p2 = _match_dims(poly1, poly2)
    if _matching_internal_domain(p1, p2):
        l2n_p1 = minterpy.transformations.LagrangeToNewton(p1)
        newt_p1 = l2n_p1()
        l2n_p2 = minterpy.transformations.LagrangeToNewton(p2)
        newt_p2 = l2n_p2()

        max_poly_degree = np.max(
            np.array([p1.multi_index.poly_degree, p2.multi_index.poly_degree])
        )
        max_lp_degree = np.max(
            np.array([p1.multi_index.lp_degree, p2.multi_index.lp_degree])
        )

        dim = p1.spatial_dimension  # must be the same for p2

        res_mi = MultiIndexSet.from_degree(dim, int(max_poly_degree), max_lp_degree)
        res_grid = Grid(res_mi)

        un = res_grid.unisolvent_nodes

        eval_p1 = newt_p1(un)
        eval_p2 = newt_p2(un)

        res_coeffs = eval_p1 - eval_p2

        return LagrangePolynomial(
            res_mi,
            res_coeffs,
            grid=res_grid,
            internal_domain=p1.internal_domain,
            user_domain=p1.user_domain,
        )
    else:
        raise NotImplementedError(
            "Subtraction is not implemented for lagrange polynomials with different domains"
        )


def _lagrange_mul(
    poly1: MultivariatePolynomialSingleABC, poly2: MultivariatePolynomialSingleABC
) -> MultivariatePolynomialSingleABC:
    """Multiplication of two polynomials in Lagrange basis.


    :param poly1: First polynomial to be multiplied.
    :type poly1: LagrangePolynomial
    :param poly2: Second polynomial to be multiplied.
    :type poly2: LagrangePolynomial

    :return: The result of ``poly1 * poly2``
    :rtype: LagrangePolynomial
    """
    p1, p2 = _match_dims(poly1, poly2)
    if _matching_internal_domain(p1, p2):
        l2n_p1 = minterpy.transformations.LagrangeToNewton(p1)
        newt_p1 = l2n_p1()
        l2n_p2 = minterpy.transformations.LagrangeToNewton(p2)
        newt_p2 = l2n_p2()

        degree_poly1 = p1.multi_index.poly_degree
        degree_poly2 = p2.multi_index.poly_degree
        lpdegree_poly1 = p1.multi_index.lp_degree
        lpdegree_poly2 = p2.multi_index.lp_degree

        res_degree = int(degree_poly1 + degree_poly2)
        res_lpdegree = lpdegree_poly1 + lpdegree_poly2

        res_mi = MultiIndexSet.from_degree(
            p1.spatial_dimension, res_degree, res_lpdegree
        )
        res_grid = Grid(res_mi)

        un = res_grid.unisolvent_nodes

        eval_p1 = newt_p1(un)
        eval_p2 = newt_p2(un)

        res_coeffs = eval_p1 * eval_p2

        return LagrangePolynomial(
            res_mi,
            res_coeffs,
            grid=res_grid,
            internal_domain=p1.internal_domain,
            user_domain=p1.user_domain,
        )
    else:
        raise NotImplementedError(
            "Multiplication is not implemented for Lagrange polynomials with different domains"
        )


# TODO redundant
lagrange_generate_internal_domain = verify_domain
lagrange_generate_user_domain = verify_domain


class LagrangePolynomial(MultivariatePolynomialSingleABC):
    """Datatype to discribe polynomials in Lagrange base.

    A polynomial in Lagrange basis is the sum of so called Lagrange polynomials (each multiplied with a coefficient).
    A `single` Lagrange monomial is per definition 1 on one of the grid points and 0 on all others.

    Attributes
    ----------
    coeffs
    nr_active_monomials
    spatial_dimension
    unisolvent_nodes

    Notes
    -----
    A polynomial in Lagrange basis is well defined also for multi indices which are lexicographically incomplete. This means that the corresponding Lagrange polynomials also form a basis in such cases. These Lagrange polynomials however will possess their special property of being 1 on a single grid point and 0 on all others, with respect to the given grid! This allows defining very "sparse" polynomials (few multi indices -> few coefficients), but which still fulfill additional constraints (vanish on additional grid points). Practically this can be achieved by storing a "larger" grid (defined on a larger set of multi indices). In this case the transformation matrices become non-square, since there are fewer Lagrange polynomials than there are grid points (<-> only some of the Lagrange polynomials of this basis are "active"). Conceptually this is equal to fix the "inactivate" coefficients to always be 0.

    .. todo::
        - provide a short definition of this base here.
    """

    _newt_coeffs_lagr_monomials: Optional[ARRAY] = None

    # Virtual Functions
    _add = staticmethod(_lagrange_add)
    _sub = staticmethod(_lagrange_sub)
    _mul = staticmethod(_lagrange_mul)
    _div = staticmethod(dummy)  # type: ignore
    _pow = staticmethod(dummy)  # type: ignore
    _eval = staticmethod(dummy)  # type: ignore

    _partial_diff = staticmethod(dummy)
    _diff = staticmethod(dummy)

    generate_internal_domain = staticmethod(lagrange_generate_internal_domain)
    generate_user_domain = staticmethod(lagrange_generate_user_domain)



#!/usr/bin/env python
""" functions for input verification
"""

from typing import Optional, Sized

import numpy as np
from _warnings import warn

from minterpy.global_settings import DEBUG, DEFAULT_DOMAIN, FLOAT_DTYPE, INT_DTYPE

__author__ = "Jannik Michelfeit"
__copyright__ = "Copyright 2021, minterpy"
__credits__ = ["Jannik Michelfeit"]
# __license__ =
# __version__ =
# __maintainer__ =
__email__ = "jannik@michelfe.it"
__status__ = "Development"


def verify_domain(domain, spatial_dimension):
    """Building and verification of domains.

    This function builds a suitable domain as the cartesian product of a one-
    dimensional domain, or verifies the domain shape, of a multivariate domain is
    passed. If None is passed, the default domain is build from [-1,1].

    :param domain: Either one-dimensional domain ``(min,max)``, or a stack of domains for each domain with shape ``(spatial_dimension,2)``. If :class:`None` is passed, the ``DEFAULT_DOMAIN`` is repeated for each spatial dimentsion.
    :type domain: array_like, None
    :param spatial_dimension: Dimentsion of the domain space.
    :type spatial_dimension: int

    :return verified_domain: Stack of domains for each dimension with shape ``(spatial_dimension,2)``.
    :rtype: np.ndarray
    :raise ValueError: If no domain with the expected shape can be constructed from the input.

    """
    if domain is None:
        domain = np.repeat([DEFAULT_DOMAIN], spatial_dimension, axis=0)
    domain = np.require(domain, dtype=FLOAT_DTYPE)
    if domain.ndim == 1:
        domain = np.repeat([domain], spatial_dimension, axis=0)
    check_shape(domain, shape=(spatial_dimension, 2))
    return domain


def rectify_query_points(x, m):
    """Rectify input arguments.

    This function checks if a given input has the correct shape, or if the correct shape can be infered. For the latter it returns a version of the input with the correct shape. Correct shape means here ``(N,m)``, where ``N`` is the number of points and ``m`` the dimentsion of the domain space.

    :param x: Array of arguemnts passed to a function.
    :type x: np.ndarray

    :param m: Dimension of the domain space.
    :type m: int

    :raise ValueError: If the input array has not the expected dimensionality.

    :return: ``(nr_points,x)`` where ``nr_points`` is the number of points and ``x`` is a version of the input array with the correct shape. If ``x`` already had the right shape, it is passed without copying.
    :rtype: tuple(int,np.ndarray)

    """
    # TODO simplify. always require an unambiguous input shape!
    query_point_shape = x.shape
    if x.ndim == 1:
        if m == 1:  # -> every entry is a query point
            nr_points = query_point_shape[0]
        else:  # m > 1, interpret input as a single point -> dimensions must match!
            if len(x) != m:
                raise ValueError(
                    f"points x given as vector of shape {query_point_shape} (1D). "
                    f"detected dimensionality of the exponents however is {m}"
                )
            nr_points = 1
        x = x.reshape(nr_points, m)  # reshape to 2D
    else:
        nr_points, m_points = x.shape
        if m != m_points:
            raise ValueError(
                f"the dimensionality of the input points {m_points} "
                f"does not match the polynomial dimensionality {m}"
            )
    return nr_points, x


def rectify_eval_input(x, coefficients, exponents, verify_input):
    """Rectify input for evaluation.

    .. todo::
        - refac this based on the respective datatypes, e.g. :class:`MultiIndex` etc.
        - function signature if somewhat misleading.

    :param x:
    :type x: np.ndarray
    :param coefficients: The coefficients of the Newton polynomials. Note, format fixed such that 'lagrange2newton' conversion matrices can be passed as the Newton coefficients of all Lagrange monomials of a polynomial without prior transponation
    :type coefficients: np.ndarray
    :param exponents: a multi index ``alpha`` for every Newton polynomial corresponding to the exponents of this ``monomial``
    :type exponents: np.ndarray
    :param verify_input: weather the data types of the input should be checked.
    :type verify_input: bool

    :raise ValueError: If the number of coefficients does not match the number of monomials.
    :raise TypeError: if the input hasn't the expected dtype.

    :return: ``( N, coefficients, m, nr_points, nr_polynomials, x)`` with the rectifyed versions of the ``coefficients`` and the input array ``x``. Furthermore, return the number of exponents ``N``, the dimentsion of the domain space ``m``, the number of passed points ``nr_points`` and the number of polynomials ``nr_polynomials``.
    :rtype: (int, np.ndarray, int, int, int, np.ndarray)
    """
    N, m = exponents.shape
    if N == 0:
        raise ValueError("at least 1 monomial must be given")
    # NOTE: silently reshaping the input is dangerous,
    # because the value order could be changed without being noticed by the user -> unexpected results!
    nr_points, x = rectify_query_points(x, m)
    coeff_shape = coefficients.shape
    if len(coeff_shape) == 1:  # 1D: a single query polynomial
        nr_polynomials = 1
        coefficients = coefficients.reshape(N, nr_polynomials)  # reshape to 2D
    else:
        N_coeffs, nr_polynomials = coefficients.shape
        if N != N_coeffs:
            raise ValueError(
                f"the coefficient amount {N_coeffs} does not match the amount of monomials {N}"
            )

    if verify_input:
        expected_types = {
            "x": (x, FLOAT_DTYPE),
            "coefficients": (coefficients, FLOAT_DTYPE),
            "exponents": (exponents, INT_DTYPE),
        }
        for name, (value, exp_type) in expected_types.items():
            if value.dtype != exp_type:
                raise TypeError(
                    f"expected dtype {exp_type} for {name} but the dtype is {value.dtype}"
                )

    return N, coefficients, m, nr_points, nr_polynomials, x


def convert_eval_output(results_placeholder):
    """Converts an array to its squeezed version.

    The input array is copyed if necessary and the result is at least 1D.

    :param results_placeholder: Array to be converted.
    :type results_placeholder: np.ndarray

    :return: A (at least 1D) squeezed version of the input array.
    :rtype: np.ndarray

    .. todo::
        - use ``np.atleast_1D`` instead of the size check.


    """
    # TODO
    # convert into the expected shape
    out = results_placeholder.squeeze()
    if out.size == 1:  # do not return 0D but rather 1D array:
        out = out.reshape(1)
    return out


def check_type(a, expected_type=np.ndarray, *args, **kwargs):
    """Verify if the input has the expected type.

    :param a: Object to be checked.
    :type a: Any

    :param expected_type: The type which is check against. Default is ``np.ndarray``
    :type expected_type: type, optional

    :raise TypeError: if input object hasn't the expected type.

    .. todo::
        - why not use ``isinstance``?
        - why pass ``*args, **kwargs``?

    """
    if not issubclass(type(a), expected_type):
        raise TypeError(
            f"input must be given as {expected_type} (encountered {type(a)})"
        )


def check_dtype(a: np.ndarray, expected_dtype):
    """Verify if the input array has the expected dtype.

    :param a: Array to be checked.
    :type a: np.ndarray

    :param expected_dtype: The dtype which is check against.
    :type expected_dtype: type

    :raise TypeError: if input array hasn't the expected dtype.

    .. todo::
        - use ``is not`` instead of ``!=``.

    """
    if a.dtype != expected_dtype:
        raise TypeError(
            f"input must be given as {expected_dtype} (encountered {a.dtype})"
        )


def check_values(a: np.ndarray, *args, **kwargs):
    """Verify that the input array has neither ``NaN`` nor ``inf`` values.

    :param a: Array to be checked.
    :type a: np.ndarray

    :raise ValueError: if input array contains either ``NaN`` or ``inf`` values (or both).

    .. todo::
        - why pass ``*args, **kwargs``?

    """
    if np.any(np.isnan(a)) or np.any(np.isinf(a)):
        raise ValueError("values must not be NaN or infinity!")


def check_type_n_values(a: np.ndarray, *args, **kwargs):
    """Verify that the input array has correct type and does neither contain ``NaN`` nor ``inf`` values.

    :param a: Array to be checked.
    :type a: np.ndarray

    .. todo::
        - why pass ``*args, **kwargs``?

    See Also
    --------
    check_type : verification of the type
    check_values : verification of the values

    """
    check_type(a, *args, **kwargs)
    check_values(a, *args, **kwargs)


def check_shape(
    a: np.ndarray, shape: Optional[Sized] = None, dimensionality: int = None
):
    """Verify the shape of an input array.


    :param a: array to be checked.
    :type a: np.ndarray
    :param shape: the expected shape.Note, non integer values will be interpreted as variable size in the respective dimension. Default is :class:`None`.
    :type shape: {None,tuple,list}
    :param dimensionality: dimension of the domain space (right?). Default is :class:`None`
    :type dimensionality: {None,int}

    :raise ValueError: If input array hasn't the expected dimensionality.
    :raise ValueError: If input array hasn't the expected size.

    """
    if shape is None:
        return
    if dimensionality is None:  # check for the dimensionality by the given shape:
        dimensionality = len(shape)
    if a.ndim != dimensionality:
        raise ValueError(
            f"expected {dimensionality}D array, but encountered array of dimensionality {len(a.shape)}"
        )

    for dim, (true_size, expected_size) in enumerate(zip(a.shape, shape)):  # type: ignore
        if isinstance(expected_size, int) and true_size != expected_size:
            raise ValueError(
                f"expected array of size {expected_size} in dimension {dim},"
                f" but encountered size {true_size}"
            )


DOMAIN_WARN_MSG2 = "the grid points must fit the interpolation domain [-1;1]^m."
DOMAIN_WARN_MSG = (
    "this may lead to unexpected behaviour, "
    "e.g. rank deficiencies in the regression matrices, etc. ."
)


def check_domain_fit(points: np.ndarray):
    """Checks weather a given array of points is properly formatted and spans the standard domain :math:`[-1,1]^m`.

    .. todo::
        - maybe remove the warnings.
        - generalise to custom ``internal_domain``

    :param points: array to be checked. Here ``m`` is the dimenstion of the domain and ``k`` is the number of points.
    :type points: np.ndarray, shape = (m, k)
    :raises ValueError: if the grid points do not fit into the domain :math:`[-1;1]^m`.
    :raises ValueError: if less than one point is passed.

    """
    # check first if the sample points are valid
    check_type_n_values(points)
    # check weather the points lie outside of the domain
    sample_max = np.max(points, axis=1)
    if not np.allclose(np.maximum(sample_max, 1.0), 1.0):
        raise ValueError(DOMAIN_WARN_MSG2 + f"violated max: {sample_max}")
    sample_min = np.min(points, axis=1)
    if not np.allclose(np.minimum(sample_min, -1.0), -1.0):
        raise ValueError(DOMAIN_WARN_MSG2 + f"violated min: {sample_min}")
    check_shape(points, dimensionality=2)
    nr_of_points, m = points.shape
    if nr_of_points == 0:
        raise ValueError("at least one point must be given")
    if nr_of_points == 1:
        return  # one point cannot span the domain
    if DEBUG:
        # check weather the points span the hole domain
        max_grid_val = np.max(sample_max)
        if not np.isclose(max_grid_val, 1.0):
            warn(
                f"the highest encountered value in the given points is {max_grid_val}  (expected 1.0). "
                + DOMAIN_WARN_MSG
            )
        min_grid_val = np.min(sample_min)
        if not np.isclose(min_grid_val, -1.0):
            warn(
                f"the smallest encountered value in the given points is {min_grid_val} (expected -1.0). "
                + DOMAIN_WARN_MSG
            )



import math
import numpy as np
import pytest

from scipy.sparse.linalg import gmres
from numpy.testing import assert_

import minterpy as mp
from minterpy.extras.regression import OrdinaryRegression
from minterpy.polynomials import (
    LagrangePolynomial, NewtonPolynomial, CanonicalPolynomial
)

from conftest import (
    SpatialDimension,
    LpDegree,
    PolyDegree,
    assert_call,
    assert_multi_index_equal,
    assert_grid_equal,
    build_random_newton_polynom,
)


def _solve_gmres(rr: np.ndarray, yy: np.ndarray, ww: np.ndarray, **kwargs):
    """Solve a least-squares problem using a Generalized Min. Resid. solver."""
    if ww is None:
        coeffs, _ = gmres(rr, yy, atol="legacy", **kwargs)
    else:
        coeffs, _ = gmres(
            rr.T @ ww @ rr, rr.T @ ww @ yy, atol="legacy", **kwargs
        )

    return coeffs


# Fixtures for least-squares solver keyword arguments (These are the defaults)
LSQ_SOLVER_ARGS = {
    "lstsq": {"cond": None, "check_finite": True},
    "pinv": {"rcond": 1e-15},
    "dgesv": {"overwrite_a": False},
    "dsysv": {"lower": False},
    "dposv": {"overwrite_b": False},
}


def assert_uninitialized_polynomial_equal(poly_1, poly_2):
    try:
        assert_(isinstance(poly_1, type(poly_2)))
        assert_multi_index_equal(poly_1.multi_index, poly_2.multi_index)
        with pytest.raises(ValueError):
            _ = poly_1.coeffs
            _ = poly_2.coeffs
    except AssertionError as a:
        raise AssertionError(
            f"The two instances of {poly_1.__class__.__name__} "
            f"are not equal:\n\n {a}"
        )


# Fixtures for least-squares solver
least_squares_solvers = [
    "lstsq",
    "inv",
    "pinv",
    "dgesv",
    "dsysv",
    "dposv",
    "qr",
    "svd",
    _solve_gmres,
]


@pytest.fixture(params=least_squares_solvers)
def least_squares_solver(request):
    return request.param


# Fixtures for least-squares solver
origin_polys = [
    LagrangePolynomial, NewtonPolynomial, CanonicalPolynomial
]


@pytest.fixture(params=origin_polys)
def origin_poly(request):
    return request.param


def test_ordinary_regression_init_multi_index(
        SpatialDimension, PolyDegree, LpDegree
):
    """Test the initialization."""
    # Create a multi-index set
    multi_index_ref = mp.MultiIndexSet.from_degree(
        SpatialDimension, PolyDegree, LpDegree
    )

    # Create a grid from the multi-index set
    grid_ref = mp.Grid(multi_index_ref)

    # Create an instance of origin poly
    origin_poly_ref = mp.LagrangePolynomial(multi_index_ref)

    # Create an instance using obligatory parameters
    my_ordinary_regression = OrdinaryRegression(multi_index_ref)

    # Multi-index attribute is correctly assigned
    assert_multi_index_equal(
        my_ordinary_regression.multi_index, multi_index_ref
    )

    # Grid attribute is correctly assigned
    assert_grid_equal(my_ordinary_regression.grid, grid_ref)

    # Origin polynomial is correctly instantiated
    assert_uninitialized_polynomial_equal(
        my_ordinary_regression.origin_poly, origin_poly_ref
    )

    # Assert uninitialized attributes prior to fitting
    assert my_ordinary_regression.loocv_error is None
    assert my_ordinary_regression.regfit_l1_error is None
    assert my_ordinary_regression.regfit_l2_error is None
    assert my_ordinary_regression.eval_poly is None

    # show() method can be called
    assert_call(my_ordinary_regression.show)


def test_ordinary_regression_init_grid(
        SpatialDimension, PolyDegree, LpDegree, origin_poly
):
    """Test the initialization."""
    # Create a multi-index set
    multi_index_ref = mp.MultiIndexSet.from_degree(
        SpatialDimension, PolyDegree, LpDegree
    )

    # Create a grid from the multi-index set
    grid_ref = mp.Grid(multi_index_ref)

    # Initializer can be called only with a grid instance
    my_ordinary_regression = OrdinaryRegression(grid=grid_ref)

    # Multi-index attribute is correctly assigned
    assert_multi_index_equal(
        my_ordinary_regression.multi_index, multi_index_ref
    )

    # Grid attribute is correctly assigned
    assert_grid_equal(my_ordinary_regression.grid, grid_ref)

    # Assert uninitialized attributes prior to fitting
    assert my_ordinary_regression.loocv_error is None
    assert my_ordinary_regression.regfit_l1_error is None
    assert my_ordinary_regression.regfit_l2_error is None
    assert my_ordinary_regression.eval_poly is None


def test_ordinary_regression_init_grid_multi_index(
    SpatialDimension,
    PolyDegree,
    LpDegree,
    origin_poly,
):
    """Test the initialization."""
    # Create a multi-index set
    multi_index_ref = mp.MultiIndexSet.from_degree(
        SpatialDimension, PolyDegree, LpDegree
    )

    # Create a grid from the multi-index set
    grid_ref = mp.Grid(multi_index_ref)

    # Create an instance of origin polynomial
    origin_poly_ref = origin_poly(multi_index=multi_index_ref, grid=grid_ref)

    # Initializer can be called with different origin poly basis
    my_ordinary_regression = OrdinaryRegression(
        multi_index_ref, grid=grid_ref, origin_poly=origin_poly
    )

    # Multi-index attribute is correctly assigned
    assert_multi_index_equal(
        my_ordinary_regression.multi_index, multi_index_ref
    )

    # Grid attribute is correctly assigned
    assert_grid_equal(my_ordinary_regression.grid, grid_ref)

    # Origin polynomial is correctly instantiated
    assert_uninitialized_polynomial_equal(
        my_ordinary_regression.origin_poly, origin_poly_ref
    )

    # Assert uninitialized attributes prior to fitting
    assert my_ordinary_regression.loocv_error is None
    assert my_ordinary_regression.regfit_l1_error is None
    assert my_ordinary_regression.regfit_l2_error is None
    assert my_ordinary_regression.eval_poly is None


def test_ordinary_regression_fit(
    SpatialDimension,
    PolyDegree,
    LpDegree,
    least_squares_solver,
    origin_poly,
):
    """Test if a fit can be carried out on the instance."""
    # Create a multi-index set
    multi_index = mp.MultiIndexSet.from_degree(
        SpatialDimension, PolyDegree, LpDegree
    )

    # Create a test training set from a random Newton polynomial
    newton_poly = build_random_newton_polynom(
        SpatialDimension, PolyDegree, LpDegree
    )
    xx_train = newton_poly.grid.unisolvent_nodes
    yy_train = newton_poly(xx_train)

    # Create an OrdinaryRegression instance
    my_ordinary_regression = OrdinaryRegression(
        multi_index, origin_poly=origin_poly
    )

    # fit() can be called
    # NOTE: Regression on the unisolvent nodes with training points at the
    #       unisolvent nodes is interpolation
    my_ordinary_regression.fit(xx_train, yy_train)
    assert np.allclose(yy_train, my_ordinary_regression(xx_train))

    # Test fit() call with a vector of weights
    weights = np.ones(yy_train.shape)
    assert_call(my_ordinary_regression.fit, xx_train, yy_train, weights=weights)

    # Test fit() call with a matrix of weights
    weights = np.eye(yy_train.shape[0])
    assert_call(my_ordinary_regression.fit, xx_train, yy_train, weights=weights)

    # Different least-squares solver
    assert_call(
        my_ordinary_regression.fit,
        xx_train,
        yy_train,
        weights=weights,
        lstsq_solver=least_squares_solver
    )

    # fit() can be called with additional fitting arguments
    kwargs = LSQ_SOLVER_ARGS.get(least_squares_solver, {})
    assert_call(
        my_ordinary_regression.fit,
        xx_train,
        yy_train,
        lstsq_solver=least_squares_solver,
        **kwargs
    )

    # Assert uninitialized attributes prior to fitting
    assert my_ordinary_regression.loocv_error is not None
    assert my_ordinary_regression.regfit_l1_error is not None
    assert my_ordinary_regression.regfit_l2_error is not None
    assert my_ordinary_regression.eval_poly is not None

    # show() method can be called
    assert_call(my_ordinary_regression.show)


def test_ordinary_regression_fit_not_implemented_solver(
    SpatialDimension,
    PolyDegree,
    LpDegree,
):
    """Test if a fit with not implemented solver raises an error."""
    # Create a multi-index set
    multi_index = mp.MultiIndexSet.from_degree(
        SpatialDimension, PolyDegree, LpDegree
    )

    # Create a test training set from a random Newton polynomial
    newton_poly = build_random_newton_polynom(
        SpatialDimension, PolyDegree, LpDegree
    )
    xx_train = newton_poly.grid.unisolvent_nodes
    yy_train = newton_poly(xx_train)

    # Create an OrdinaryRegression instance
    my_ordinary_regression = OrdinaryRegression(multi_index)

    with pytest.raises(NotImplementedError):
        my_ordinary_regression.fit(xx_train, yy_train, lstsq_solver="rand123")


def test_ordinary_regression_predict(
    SpatialDimension,
    PolyDegree,
    LpDegree,
    origin_poly,
):
    """Test if the prediction can be carried out on the instance."""
    # Create a multi-index set
    multi_index = mp.MultiIndexSet.from_degree(
        SpatialDimension, PolyDegree, LpDegree
    )

    # Create the corresponding grid
    grid = mp.Grid(multi_index)

    # Create a test training set from a random Newton polynomial
    newton_poly = build_random_newton_polynom(
        SpatialDimension, PolyDegree, LpDegree
    )
    xx_train = newton_poly.grid.unisolvent_nodes
    yy_train = newton_poly(xx_train)

    # Create and fit an OrdinaryRegression instance
    my_ordinary_regression = OrdinaryRegression(
        multi_index=multi_index, grid=grid, origin_poly=origin_poly
    )

    # predict() without fitting
    with pytest.raises(TypeError):
        my_ordinary_regression.predict(xx_train)  # predict method
        my_ordinary_regression(xx_train)          # __call__ dunder method

    # Fit the ordinary regression model
    my_ordinary_regression.fit(xx_train, yy_train)

    # predict() can be called
    assert_call(my_ordinary_regression.predict, xx_train)  # predict method
    assert_call(my_ordinary_regression, xx_train)          # __call__ method

    # Test the prediction on a test dataset
    xx_test = -1 + 2 * np.random.rand(10, SpatialDimension)
    yy_test = newton_poly(xx_test)
    assert np.allclose(yy_test, my_ordinary_regression(xx_test))


def test_wrong_multi_index_instance():
    """Test if passing a wrong instance of multi-index raises an error."""

    mi = mp.MultiIndexSet.from_degree(1, 3, 1)
    mi_exponents = mi.exponents

    with pytest.raises(TypeError):
        OrdinaryRegression(mi_exponents)


def test_wrong_grid_instance():
    """Test if passing a wrong instance of grid raises an error."""

    mi = mp.MultiIndexSet.from_degree(2, 3, 1)

    with pytest.raises(TypeError):
        OrdinaryRegression(grid=mi)


def test_insufficient_input():
    """Test if passing an insufficient input to constructor raises an error."""

    with pytest.raises(ValueError):
        OrdinaryRegression()


def test_grid_subset_of_multi_index():
    """Test if passing a grid of lower degree than multi-index raises an error.
    """

    # Grid is of higher degree than the polynomial multi-index: okay
    mi_poly = mp.MultiIndexSet.from_degree(2, 3, 1)
    mi_grid = mp.MultiIndexSet.from_degree(2, 5, 1)
    grid = mp.Grid(mi_grid)

    assert_call(OrdinaryRegression, multi_index=mi_poly, grid=grid)

    # Grid is of lower degree than the polynomial multi-index: not okay
    mi_grid = mp.MultiIndexSet.from_degree(2, 2, 1)
    grid = mp.Grid(mi_grid)

    with pytest.raises(ValueError):
        OrdinaryRegression(mi_poly, grid)


def test_underdetermined_system(SpatialDimension, PolyDegree, LpDegree):
    """Test fitting under-determined system."""

    # Create a multi-index set
    multi_index = mp.MultiIndexSet.from_degree(
        SpatialDimension, PolyDegree, LpDegree
    )

    # Create a test training set from a random Newton polynomial
    newton_poly = build_random_newton_polynom(
        SpatialDimension, PolyDegree, LpDegree
    )

    # At least one training point
    sample_size = max(1, math.floor(0.8 * len(multi_index)))
    xx_train = -1 + 2 * np.random.rand(sample_size, SpatialDimension)
    yy_train = newton_poly(xx_train)

    # Create and fit an OrdinaryRegression instance
    my_ordinary_regression = OrdinaryRegression(multi_index)

    # Fit the ordinary regression model
    my_ordinary_regression.fit(xx_train, yy_train, lstsq_solver="qr")

    assert my_ordinary_regression.loocv_error == (np.infty, np.infty)


def test_single_monomial(origin_poly):
    """Test an edge case of a single monomial."""

    spatial_dimension = 1
    poly_degree = 0
    lp_degree = 1
    # Create a multi-index set
    multi_index = mp.MultiIndexSet.from_degree(
        spatial_dimension, poly_degree, lp_degree
    )

    # Create a test training set from a random Newton polynomial
    newton_poly = build_random_newton_polynom(
        spatial_dimension, poly_degree, lp_degree
    )
    xx_train = -1 + 2 * np.random.rand(100, 1)
    yy_train = newton_poly(xx_train)

    # Create and fit an OrdinaryRegression instance
    my_ordinary_regression = OrdinaryRegression(
        multi_index=multi_index, origin_poly=origin_poly
    )

    # predict() without fitting
    with pytest.raises(TypeError):
        my_ordinary_regression.predict(xx_train)  # predict method
        my_ordinary_regression(xx_train)  # __call__ dunder method

    # Fit the ordinary regression model
    my_ordinary_regression.fit(xx_train, yy_train)

    # predict() can be called
    assert_call(my_ordinary_regression.predict, xx_train)  # predict method
    assert_call(my_ordinary_regression, xx_train)  # __call__ method

    # Test the prediction on a test dataset
    xx_test = -1 + 2 * np.random.rand(1000, spatial_dimension)
    yy_test = newton_poly(xx_test)
    assert np.allclose(yy_test, my_ordinary_regression(xx_test))


def test_unsupported_polynomial_basis():
    """Test fitting an unsupported polynomial basis."""
    spatial_dimension = 1
    poly_degree = 3
    lp_degree = 2.0
    # Create a multi-index set
    multi_index = mp.MultiIndexSet.from_degree(
        spatial_dimension, poly_degree, lp_degree
    )

    # Create a test training set
    xx_train = -1 + 2 * np.random.rand(100, 1)
    yy_train = 2 * xx_train

    # Create and fit an OrdinaryRegression instance
    my_ordinary_regression = OrdinaryRegression(
        multi_index=multi_index, origin_poly=_EmptyClass
    )

    with pytest.raises(TypeError):
        my_ordinary_regression.fit(xx_train, yy_train)


class _EmptyClass:
    def __init__(self, multi_index, grid):
        self.multi_index = multi_index
        self.grid = grid
        self.exponents = None
        self.generating_points = None



"""
This is the minterpy package init.

isort:skip_file
"""

from .version import version as __version__

__all__ = [
    "__version__",
]

from . import core  # noqa
from .core import *  # noqa

__all__ += core.__all__


from . import polynomials  # noqa
from .polynomials import *  # noqa

__all__ += polynomials.__all__


from . import transformations  # noqa
from .transformations import *  # noqa

__all__ += transformations.__all__


from . import interpolation  # noqa
from .interpolation import *  # noqa

__all__ += interpolation.__all__

from . import extras
from .extras import regression
from .extras.regression import *

__all__ += regression.__all__
